# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2018 Jennifer Wu
#
# Licensed under the terms of the BSD2 License
# See LICENSE.txt for details
# -----------------------------------------------------------------------------
from pyswmm.toolkitapi import LidLayers, LidLayersProperty


class Surface(object):
    """
    Methods and properties of the surface layer associated with an LID

    +--------------------+--------------------+--------------------+--------------------+
    | Layer              | Parameter          | Setter Before Sim  | Setter During Sim  |
    +====================+====================+====================+====================+
    | Surface            | thickness          | enabled            | disabled           |
    +--------------------+--------------------+--------------------+--------------------+
    | Surface            | void_fraction      | enabled            | disabled           |
    +--------------------+--------------------+--------------------+--------------------+
    | Surface            | roughness          | enabled            | enabled            |
    +--------------------+--------------------+--------------------+--------------------+
    | Surface            | slope              | enabled            | disabled           |
    +--------------------+--------------------+--------------------+--------------------+
    | Surface            | side_slope         | enabled            | disabled           |
    +--------------------+--------------------+--------------------+--------------------+
    | Surface            | alpha              | enabled            | disabled           |
    +--------------------+--------------------+--------------------+--------------------+

    .. code-block:: python

        from pyswmm import Simulation, LidControls

        with Simulation('lid_model.inp') as sim:
            lid_control = LidControls(sim)['LID_C1']

            # Handles to get/set params for each layer
            lid_control_surface = lid_control.surface

            print(lid_control_surface.roughness)
            lid_control_surface.roughness = 0.5
            print(lid_control_surface.roughness)

    .. code-block::

        >>> 0.03
        >>> 0.5

    """

    def __init__(self, model, lidcontrol):
        self._model = model
        self._lidcontrol = lidcontrol
        self._lidcontrolid = lidcontrol._lidcontrolid

    @property
    def thickness(self):
        """
        Get lid control surface layer thickness

        :return: Parameter Value
        :rtype: double
        """
        return self._model.getLidCParam(
            self._lidcontrolid,
            LidLayers.surface.value,
            LidLayersProperty.thickness.value,
        )

    @thickness.setter
    def thickness(self, param):
        """Set lid control surface layer thickness"""
        return self._model.setLidCParam(
            self._lidcontrolid,
            LidLayers.surface.value,
            LidLayersProperty.thickness.value,
            param,
        )

    @property
    def void_fraction(self):
        """
        Get lid control surface layer available fraction of storage volume

        :return: Parameter Value
        :rtype: double
        """
        return self._model.getLidCParam(
            self._lidcontrolid,
            LidLayers.surface.value,
            LidLayersProperty.voidFrac.value,
        )

    @void_fraction.setter
    def void_fraction(self, param):
        """Set lid control surface layer available fraction of storage volume"""
        return self._model.setLidCParam(
            self._lidcontrolid,
            LidLayers.surface.value,
            LidLayersProperty.voidFrac.value,
            param,
        )

    @property
    def roughness(self):
        """
        Get lid control surface layer surface Mannings n

        :return: Parameter Value
        :rtype: double
        """
        return self._model.getLidCParam(
            self._lidcontrolid,
            LidLayers.surface.value,
            LidLayersProperty.roughness.value,
        )

    @roughness.setter
    def roughness(self, param):
        """Set lid control surface layer surface Mannings n"""
        return self._model.setLidCParam(
            self._lidcontrolid,
            LidLayers.surface.value,
            LidLayersProperty.roughness.value,
            param,
        )

    @property
    def slope(self):
        """
        Get lid control surface layer land surface slope (fraction)

        :return: Parameter Value
        :rtype: double
        """
        return self._model.getLidCParam(
            self._lidcontrolid,
            LidLayers.surface.value,
            LidLayersProperty.surfSlope.value,
        )

    @slope.setter
    def slope(self, param):
        """Set lid control surface layer land surface slope (fraction)"""
        return self._model.setLidCParam(
            self._lidcontrolid,
            LidLayers.surface.value,
            LidLayersProperty.surfSlope.value,
            param,
        )

    @property
    def side_slope(self):
        """
        Get lid control surface layer swale side slope (run/rise)

        :return: Parameter Value
        :rtype: double
        """
        return self._model.getLidCParam(
            self._lidcontrolid,
            LidLayers.surface.value,
            LidLayersProperty.sideSlope.value,
        )

    @side_slope.setter
    def side_slope(self, param):
        """Set lid control surface layer swale side slope (run/rise)"""
        return self._model.setLidCParam(
            self._lidcontrolid,
            LidLayers.surface.value,
            LidLayersProperty.sideSlope.value,
            param,
        )

    @property
    def alpha(self):
        """
        Get lid control surface layer swale side slope (run/rise)

        :return: Parameter Value
        :rtype: double
        """
        return self._model.getLidCParam(
            self._lidcontrolid, LidLayers.surface.value, LidLayersProperty.alpha.value
        )


class Soil(object):
    """
    Methods and properties of the soil layer associated with an LID

    +--------------------+--------------------+--------------------+--------------------+
    | Layer              | Parameter          | Setter Before Sim  | Setter During Sim  |
    +====================+====================+====================+====================+
    | Soil               | thickness          | enabled            | disabled           |
    +--------------------+--------------------+--------------------+--------------------+
    | Soil               | porosity           | enabled            | disabled           |
    +--------------------+--------------------+--------------------+--------------------+
    | Soil               | field_capacity     | enabled            | disabled           |
    +--------------------+--------------------+--------------------+--------------------+
    | Soil               | wilting_point      | enabled            | disabled           |
    +--------------------+--------------------+--------------------+--------------------+
    | Soil               | k_saturated        | enabled            | disabled           |
    +--------------------+--------------------+--------------------+--------------------+
    | Soil               | k_slope            | enabled            | disabled           |
    +--------------------+--------------------+--------------------+--------------------+
    | Soil               | suction_head       | enabled            | disabled           |
    +--------------------+--------------------+--------------------+--------------------+

    .. code-block:: python

        from pyswmm import Simulation, LidControls

        with Simulation('lid_model.inp') as sim:
            lid_control = LidControls(sim)['LID_C1']

            # Handles to get/set params for each layer
            lid_control_soil = lid_control.soil

            print(lid_control_soil.porosity)
            lid_control_soil.porosity = 0.5
            print(lid_control_soil.porosity)
    """

    def __init__(self, model, lidcontrol):
        self._model = model
        self._lidcontrol = lidcontrol
        self._lidcontrolid = lidcontrol._lidcontrolid

    @property
    def thickness(self):
        """
        Get lid control soil layer thickness

        :return: Parameter Value
        :rtype: double
        """
        return self._model.getLidCParam(
            self._lidcontrolid, LidLayers.soil.value, LidLayersProperty.thickness.value
        )

    @thickness.setter
    def thickness(self, param):
        """Set lid control soil layer thickness"""
        return self._model.setLidCParam(
            self._lidcontrolid,
            LidLayers.soil.value,
            LidLayersProperty.thickness.value,
            param,
        )

    @property
    def porosity(self):
        """
        Get lid control soil layer void volume / total volume

        :return: Parameter Value
        :rtype: double
        """
        return self._model.getLidCParam(
            self._lidcontrolid, LidLayers.soil.value, LidLayersProperty.porosity.value
        )

    @porosity.setter
    def porosity(self, param):
        """Set lid control soil layer void volume / total volume"""
        return self._model.setLidCParam(
            self._lidcontrolid,
            LidLayers.soil.value,
            LidLayersProperty.porosity.value,
            param,
        )

    @property
    def field_capacity(self):
        """
        Get lid control soil layer field capacity

        :return: Parameter Value
        :rtype: double
        """
        return self._model.getLidCParam(
            self._lidcontrolid, LidLayers.soil.value, LidLayersProperty.fieldCap.value
        )

    @field_capacity.setter
    def field_capacity(self, param):
        """Set lid control soil layer field capacity"""
        return self._model.setLidCParam(
            self._lidcontrolid,
            LidLayers.soil.value,
            LidLayersProperty.fieldCap.value,
            param,
        )

    @property
    def wilting_point(self):
        """
        Get lid control soil layer wilting point

        :return: Parameter Value
        :rtype: double
        """
        return self._model.getLidCParam(
            self._lidcontrolid, LidLayers.soil.value, LidLayersProperty.wiltPoint.value
        )

    @wilting_point.setter
    def wilting_point(self, param):
        """Set lid control soil layer wilting point"""
        return self._model.setLidCParam(
            self._lidcontrolid,
            LidLayers.soil.value,
            LidLayersProperty.wiltPoint.value,
            param,
        )

    @property
    def k_saturated(self):
        """
        Get lid control soil layer saturated hydraulic conductivity

        :return: Parameter Value
        :rtype: double
        """
        return self._model.getLidCParam(
            self._lidcontrolid, LidLayers.soil.value, LidLayersProperty.kSat.value
        )

    @k_saturated.setter
    def k_saturated(self, param):
        """Set lid control soil layer saturated hydraulic conductivity"""
        return self._model.setLidCParam(
            self._lidcontrolid,
            LidLayers.soil.value,
            LidLayersProperty.kSat.value,
            param,
        )

    @property
    def k_slope(self):
        """
        Get lid control soil layer slope of log(k) v. moisture content curve

        :return: Parameter Value
        :rtype: double
        """
        return self._model.getLidCParam(
            self._lidcontrolid, LidLayers.soil.value, LidLayersProperty.kSlope.value
        )

    @k_slope.setter
    def k_slope(self, param):
        """Set lid control soil layer slope of log(k) v. moisture content curve"""
        return self._model.setLidCParam(
            self._lidcontrolid,
            LidLayers.soil.value,
            LidLayersProperty.kSlope.value,
            param,
        )

    @property
    def suction_head(self):
        """
        Get lid control soil layer suction head at wetting front

        :return: Parameter Value
        :rtype: double
        """
        return self._model.getLidCParam(
            self._lidcontrolid, LidLayers.soil.value, LidLayersProperty.suction.value
        )

    @suction_head.setter
    def suction_head(self, param):
        """Set lid control soil layer suction head at wetting front"""
        return self._model.setLidCParam(
            self._lidcontrolid,
            LidLayers.soil.value,
            LidLayersProperty.suction.value,
            param,
        )


class Storage(object):
    """
    Methods and properties of the storage layer associated with an LID

    +--------------------+--------------------+--------------------+--------------------+
    | Layer              | Parameter          | Setter Before Sim  | Setter During Sim  |
    +====================+====================+====================+====================+
    | Storage            | thickness          | enabled            | disabled           |
    +--------------------+--------------------+--------------------+--------------------+
    | Storage            | void_fraction      | enabled            | disabled           |
    +--------------------+--------------------+--------------------+--------------------+
    | Storage            | k_saturated        | enabled            | disabled           |
    +--------------------+--------------------+--------------------+--------------------+
    | Storage            | clog_factor        | enabled            | enabled            |
    +--------------------+--------------------+--------------------+--------------------+


    .. code-block:: python

        from pyswmm import Simulation, LidControls

        with Simulation('lid_model.inp') as sim:
            lid_control = LidControls(sim)['LID_C1']

            # Handles to get/set params for each layer
            lid_control_storage = lid_control.storage

            print(lid_control_storage.porosity)
            lid_control_storage.porosity = 0.5
            print(lid_control_storage.porosity)
    """

    def __init__(self, model, lidcontrol):
        self._model = model
        self._lidcontrol = lidcontrol
        self._lidcontrolid = lidcontrol._lidcontrolid

    @property
    def thickness(self):
        """
        Get lid control storage layer thickness

        :return: Parameter Value
        :rtype: double
        """
        return self._model.getLidCParam(
            self._lidcontrolid,
            LidLayers.storage.value,
            LidLayersProperty.thickness.value,
        )

    @thickness.setter
    def thickness(self, param):
        """Set lid control storage layer thickness"""
        return self._model.setLidCParam(
            self._lidcontrolid,
            LidLayers.storage.value,
            LidLayersProperty.thickness.value,
            param,
        )

    @property
    def void_fraction(self):
        """
        Get lid control storage layer void volume / total volume

        :return: Parameter Value
        :rtype: double
        """
        return self._model.getLidCParam(
            self._lidcontrolid,
            LidLayers.storage.value,
            LidLayersProperty.voidFrac.value,
        )

    @void_fraction.setter
    def void_fraction(self, param):
        """Set lid control storage layer void volume / total volume"""
        return self._model.setLidCParam(
            self._lidcontrolid,
            LidLayers.storage.value,
            LidLayersProperty.voidFrac.value,
            param,
        )

    @property
    def k_saturated(self):
        """
        Get lid control storage layer saturated hydraulic conductivity

        :return: Parameter Value
        :rtype: double
        """
        return self._model.getLidCParam(
            self._lidcontrolid, LidLayers.storage.value, LidLayersProperty.kSat.value
        )

    @k_saturated.setter
    def k_saturated(self, param):
        """Set lid control storage layer saturated hydraulic conductivity"""
        return self._model.setLidCParam(
            self._lidcontrolid,
            LidLayers.storage.value,
            LidLayersProperty.kSat.value,
            param,
        )

    @property
    def clog_factor(self):
        """
        Get lid control storage layer clogging factor

        :return: Parameter Value
        :rtype: double
        """
        return self._model.getLidCParam(
            self._lidcontrolid,
            LidLayers.storage.value,
            LidLayersProperty.clogFactor.value,
        )

    @clog_factor.setter
    def clog_factor(self, param):
        """Set lid control storage layer clogging factor"""
        return self._model.setLidCParam(
            self._lidcontrolid,
            LidLayers.storage.value,
            LidLayersProperty.clogFactor.value,
            param,
        )


class Pavement(object):
    """
    Methods and properties of the pavement layer associated with an LID

    +--------------------+--------------------+--------------------+--------------------+
    | Layer              | Parameter          | Setter Before Sim  | Setter During Sim  |
    +====================+====================+====================+====================+
    | Pavement           | thickness          | enabled            | disabled           |
    +--------------------+--------------------+--------------------+--------------------+
    | Pavement           | void_fraction      | enabled            | disabled           |
    +--------------------+--------------------+--------------------+--------------------+
    | Pavement           | impervious_fraction| enabled            | disabled           |
    +--------------------+--------------------+--------------------+--------------------+
    | Pavement           | k_saturated        | enabled            | disabled           |
    +--------------------+--------------------+--------------------+--------------------+
    | Pavement           | clog_factor        | enabled            | enabled            |
    +--------------------+--------------------+--------------------+--------------------+
    | Pavement           | regeneration       | enabled            | disabled           |
    +--------------------+--------------------+--------------------+--------------------+
    | Pavement           | regeneration_degree| enabled            | disabled           |
    +--------------------+--------------------+--------------------+--------------------+

    .. code-block:: python

        from pyswmm import Simulation, LidControls

        with Simulation('lid_model.inp') as sim:
            lid_control = LidControls(sim)['LID_C1']

            # Handles to get/set params for each layer
            lid_control_pavement = lid_control.pavement

            print(lid_control_pavement.void_fraction)
            lid_control_pavement.void_fraction = 0.05
            print(lid_control_pavement.void_fraction)

    """

    def __init__(self, model, lidcontrol):
        self._model = model
        self._lidcontrol = lidcontrol
        self._lidcontrolid = lidcontrol._lidcontrolid

    @property
    def thickness(self):
        """
        Get lid control pavement layer thickness

        :return: Parameter Value
        :rtype: double
        """
        return self._model.getLidCParam(
            self._lidcontrolid,
            LidLayers.pavement.value,
            LidLayersProperty.thickness.value,
        )

    @thickness.setter
    def thickness(self, param):
        """Get lid control pavement layer thickness"""
        return self._model.setLidCParam(
            self._lidcontrolid,
            LidLayers.pavement.value,
            LidLayersProperty.thickness.value,
            param,
        )

    @property
    def void_fraction(self):
        """
        Get lid control pavement layer void volume / total volume

        :return: Parameter Value
        :rtype: double
        """
        return self._model.getLidCParam(
            self._lidcontrolid,
            LidLayers.pavement.value,
            LidLayersProperty.voidFrac.value,
        )

    @void_fraction.setter
    def void_fraction(self, param):
        """Set lid control pavement layer void volume / total volume"""
        return self._model.setLidCParam(
            self._lidcontrolid,
            LidLayers.pavement.value,
            LidLayersProperty.voidFrac.value,
            param,
        )

    @property
    def impervious_fraction(self):
        """
        Get lid control pavement layer impervious area fraction

        :return: Parameter Value
        :rtype: double
        """
        return self._model.getLidCParam(
            self._lidcontrolid,
            LidLayers.pavement.value,
            LidLayersProperty.impervFrac.value,
        )

    @impervious_fraction.setter
    def impervious_fraction(self, param):
        """Set lid control pavement layer impervious area fraction"""
        return self._model.setLidCParam(
            self._lidcontrolid,
            LidLayers.pavement.value,
            LidLayersProperty.impervFrac.value,
            param,
        )

    @property
    def k_saturated(self):
        """
        Get lid control pavement layer permeability

        :return: Parameter Value
        :rtype: double
        """
        return self._model.getLidCParam(
            self._lidcontrolid, LidLayers.pavement.value, LidLayersProperty.kSat.value
        )

    @k_saturated.setter
    def k_saturated(self, param):
        """Get lid control pavement layer permeability"""
        return self._model.setLidCParam(
            self._lidcontrolid,
            LidLayers.pavement.value,
            LidLayersProperty.kSat.value,
            param,
        )

    @property
    def clog_factor(self):
        """
        Get lid control pavement layer clogging factor

        :return: Parameter Value
        :rtype: double
        """
        return self._model.getLidCParam(
            self._lidcontrolid,
            LidLayers.pavement.value,
            LidLayersProperty.clogFactor.value,
        )

    @clog_factor.setter
    def clog_factor(self, param):
        """Get lid control pavement layer clogging factor"""
        return self._model.setLidCParam(
            self._sim._isStarted,
            self._lidcontrolid,
            LidLayers.pavement.value,
            LidLayersProperty.clogFactor.value,
            param,
        )

    @property
    def regeneration(self):
        """
        Get lid control pavement layer clogging regeneration interval (days)

        :return: Parameter Value
        :rtype: double
        """
        return self._model.getLidCParam(
            self._lidcontrolid,
            LidLayers.pavement.value,
            LidLayersProperty.regenDays.value,
        )

    @regeneration.setter
    def regeneration(self, param):
        """Get lid control pavement layer clogging regeneration interval (days)"""
        return self._model.setLidCParam(
            self._lidcontrolid,
            LidLayers.pavement.value,
            LidLayersProperty.regenDays.value,
            param,
        )

    @property
    def regeneration_degree(self):
        """
        Get lid control pavement layer clogging regeneration degree

        :return: Parameter Value
        :rtype: double
        """
        return self._model.getLidCParam(
            self._lidcontrolid,
            LidLayers.pavement.value,
            LidLayersProperty.regenDegree.value,
        )

    @regeneration_degree.setter
    def regeneration_degree(self, param):
        """Get lid control pavement layer clogging regeneration degree"""
        return self._model.setLidCParam(
            self._lidcontrolid,
            LidLayers.pavement.value,
            LidLayersProperty.regenDegree.value,
            param,
        )


class Drain(object):
    """
    Methods and properties of the under drain layer associated with an LID

    +--------------------+--------------------+--------------------+--------------------+
    | Layer              | Parameter          | Setter Before Sim  | Setter During Sim  |
    +====================+====================+====================+====================+
    | Drain              | coefficient        | enabled            | enabled            |
    +--------------------+--------------------+--------------------+--------------------+
    | Drain              | exponent           | enabled            | enabled            |
    +--------------------+--------------------+--------------------+--------------------+
    | Drain              | offset             | enabled            | enabled            |
    +--------------------+--------------------+--------------------+--------------------+
    | Drain              | delay              | enabled            | enabled            |
    +--------------------+--------------------+--------------------+--------------------+
    | Drain              | open_head          | enabled            | enabled            |
    +--------------------+--------------------+--------------------+--------------------+
    | Drain              | close_head         | enabled            | enabled            |
    +--------------------+--------------------+--------------------+--------------------+

    .. code-block:: python

        from pyswmm import Simulation, LidControls

        with Simulation('lid_model.inp') as sim:
            lid_control = LidControls(sim)['LID_C1']

            # Handles to get/set params for each layer
            lid_control_drain = lid_control.drain

            print(lid_control_drain.coefficient)
            lid_control_drain.coefficient = 0.05
            print(lid_control_drain.coefficient)
    """

    def __init__(self, model, lidcontrol):
        self._model = model
        self._lidcontrol = lidcontrol
        self._lidcontrolid = lidcontrol._lidcontrolid

    @property
    def coefficient(self):
        """
        Get lid control drain layer underdrain flow coefficient

        :return: Parameter Value
        :rtype: double
        """
        return self._model.getLidCParam(
            self._lidcontrolid, LidLayers.drain.value, LidLayersProperty.coeff.value
        )

    @coefficient.setter
    def coefficient(self, param):
        """Set lid control drain layer underdrain flow coefficient"""
        return self._model.setLidCParam(
            self._lidcontrolid,
            LidLayers.drain.value,
            LidLayersProperty.coeff.value,
            param,
        )

    @property
    def exponent(self):
        """
        Get lid control drain layer underdrain head exponent

        :return: Parameter Value
        :rtype: double
        """
        return self._model.getLidCParam(
            self._lidcontrolid, LidLayers.drain.value, LidLayersProperty.expon.value
        )

    @exponent.setter
    def exponent(self, param):
        """Set lid control drain layer underdrain head exponent"""
        return self._model.setLidCParam(
            self._lidcontrolid,
            LidLayers.drain.value,
            LidLayersProperty.expon.value,
            param,
        )

    @property
    def offset(self):
        """
        Get lid control drain layer offset height of underdrain

        :return: Parameter Value
        :rtype: double
        """
        return self._model.getLidCParam(
            self._lidcontrolid, LidLayers.drain.value, LidLayersProperty.offset.value
        )

    @offset.setter
    def offset(self, param):
        """Set lid control drain layer offset height of underdrain"""
        return self._model.setLidCParam(
            self._lidcontrolid,
            LidLayers.drain.value,
            LidLayersProperty.offset.value,
            param,
        )

    @property
    def delay(self):
        """
        Get lid control drain layer rain barrel drain delay time (sec)

        :return: Parameter Value
        :rtype: double
        """
        return self._model.getLidCParam(
            self._lidcontrolid, LidLayers.drain.value, LidLayersProperty.delay.value
        )

    @delay.setter
    def delay(self, param):
        """Set lid control drain layer rain barrel drain delay time (sec)"""
        return self._model.setLidCParam(
            self._lidcontrolid,
            LidLayers.drain.value,
            LidLayersProperty.delay.value,
            param,
        )

    @property
    def open_head(self):
        """
        Get lid control drain layer head when drain opens (ft)

        :return: Parameter Value
        :rtype: double
        """
        return self._model.getLidCParam(
            self._lidcontrolid, LidLayers.drain.value, LidLayersProperty.hOpen.value
        )

    @open_head.setter
    def open_head(self, param):
        """Set lid control drain layer head when drain opens (ft)"""
        return self._model.setLidCParam(
            self._lidcontrolid,
            LidLayers.drain.value,
            LidLayersProperty.hOpen.value,
            param,
        )

    @property
    def close_head(self):
        """
        Get lid control drain layer head when drain closes (ft)

        :return: Parameter Value
        :rtype: double
        """
        return self._model.getLidCParam(
            self._lidcontrolid, LidLayers.drain.value, LidLayersProperty.hClose.value
        )

    @close_head.setter
    def close_head(self, param):
        """Set lid control drain layer drain closes (ft)"""
        return self._model.setLidCParam(
            self._lidcontrolid,
            LidLayers.drain.value,
            LidLayersProperty.hClose.value,
            param,
        )


class DrainMat(object):
    """
    Methods and properties of the drain mat layer associated with an LID

    +--------------------+--------------------+--------------------+--------------------+
    | Layer              | Parameter          | Setter Before Sim  | Setter During Sim  |
    +====================+====================+====================+====================+
    | DrainMat           | thickness          | enabled            | disabled           |
    +--------------------+--------------------+--------------------+--------------------+
    | DrainMat           | void_fraction      | enabled            | disabled           |
    +--------------------+--------------------+--------------------+--------------------+
    | DrainMat           | roughness          | enabled            | enabled            |
    +--------------------+--------------------+--------------------+--------------------+
    | DrainMat           | alpha              | enabled            | disabled           |
    +--------------------+--------------------+--------------------+--------------------+

    .. code-block:: python

        from pyswmm import Simulation, LidControls

        with Simulation('lid_model.inp') as sim:
            lid_control = LidControls(sim)['LID_C1']

            # Handles to get/set params for each layer
            lid_control_drain_mat = lid_control.drain_mat

            print(lid_control_drain_mat.roughness)
            lid_control_drain_mat.roughness = 0.05
            print(lid_control_drain_mat.roughness)

    """

    def __init__(self, model, lidcontrol):
        self._model = model
        self._lidcontrol = lidcontrol
        self._lidcontrolid = lidcontrol._lidcontrolid

    @property
    def thickness(self):
        """
        Get lid control drainmat layer thickness

        :return: Parameter Value
        :rtype: double
        """
        return self._model.getLidCParam(
            self._lidcontrolid,
            LidLayers.drainMat.value,
            LidLayersProperty.thickness.value,
        )

    @thickness.setter
    def thickness(self, param):
        """Set lid control drainmat layer thickness"""
        return self._model.setLidCParam(
            self._lidcontrolid,
            LidLayers.drainMat.value,
            LidLayersProperty.thickness.value,
            param,
        )

    @property
    def void_fraction(self):
        """
        Get lid control drainmat layer void volume / total volume

        :return: Parameter Value
        :rtype: double
        """
        return self._model.getLidCParam(
            self._lidcontrolid,
            LidLayers.drainMat.value,
            LidLayersProperty.voidFrac.value,
        )

    @void_fraction.setter
    def void_fraction(self, param):
        """Set lid control drainmat layer void volume / total volume"""
        return self._model.setLidCParam(
            self._lidcontrolid,
            LidLayers.drainMat.value,
            LidLayersProperty.voidFrac.value,
            param,
        )

    @property
    def roughness(self):
        """
        Get lid control drainmat layer Mannings n for green roof drainage mats

        :return: Parameter Value
        :rtype: double
        """
        return self._model.getLidCParam(
            self._lidcontrolid,
            LidLayers.drainMat.value,
            LidLayersProperty.roughness.value,
        )

    @roughness.setter
    def roughness(self, param):
        """Set lid control drainmat layer Mannings n for green roof drainage mats"""
        return self._model.setLidCParam(
            self._lidcontrolid,
            LidLayers.drainMat.value,
            LidLayersProperty.roughness.value,
            param,
        )

    @property
    def alpha(self):
        """
        Get lid control drainmat layer slope/roughness term in Manning equation

        :return: Parameter Value
        :rtype: double
        """
        return self._model.getLidCParam(
            self._lidcontrolid, LidLayers.drainMat.value, LidLayersProperty.alpha.value
        )



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2025 (see AUTHORS)
#
# Licensed under the terms of the BSD2 License
# See LICENSE.txt for details
# -----------------------------------------------------------------------------
from __future__ import annotations
from pyswmm.errors import OutputException
from datetime import datetime, timedelta
from functools import wraps
from typing import NoReturn, Optional, Union

# Third party imports
from swmm.toolkit import output, shared_enum
from julian import from_jd


def output_open_handler(func):
    """
    Checks if output file is open before running function.

    :param func : function method of Output class
    """

    @wraps(func)
    def inner_function(self, *args, **kwargs):
        if not self.loaded:
            self.open()

        return func(self, *args, **kwargs)

    return inner_function


class Output(object):
    """
    Output Methods.
    Base class for a SWMM Output binary file.
    The output object provides several options to process timeseries within output binary file.

    :param binfile: model binary file path

    Examples:

    >>> from pyswmm import Output
    >>>
    >>> with Output('tests/data/model_full_features.out') as out:
    ...     print(len(out.subcatchments))
    ...     print(len(out.nodes))
    ...     print(len(out.links))
    >>>
    >>> 3
    >>> 4
    >>> 3
    """

    def __init__(self, binfile):
        """
        Initialize the Output class.
        :param binfile: model binary file path
        """
        self.binfile = binfile

        self.handle = None
        self.loaded = False
        self.delete_handle = False
        self.period = None
        self.report = None
        self.start = None
        self.end = None
        self._times = None

        self._project_size = None
        self._subcatchments = None
        self._nodes = None
        self._links = None
        self._pollutants = None

    @staticmethod
    def verify_index(index, index_dict, index_type):
        """
        Validate the index of a model element passed to Output methods. Used to
        convert model element names to their index in the out file.

        :param index: The name or index of the model element listed in the index_dict dict
        :type index: Union[str, int]
        :param index_dict: The dict against which to validate the index
                           (one of self.nodes, self.links, self.subcatchments)
        :type index_dict: dict
        :param index_type: The type of model element (e.g. node, link, etc.)
                           Only used to print the exception if an attribute cannot be found
        :type index_type: str
        :raises OutputException: Exception if element cannot be found in dict
        :return: The integer index of the requested element
        :rtype: int
        """

        arg_index = index

        if isinstance(index, str):
            index = index_dict.get(index, None)

        if index is None:
            raise OutputException(
                f"{index_type} ID: {arg_index} does not exist in model output."
            )

        return index

    @staticmethod
    def verify_time(
        time_index: Optional[Union[datetime, int]],
        time_list: list,
        start: datetime,
        end: datetime,
        report: int,
        default_time: Union[datetime, int],
    ) -> int:
        """
        Validate time parameter passed to Output methods. Used to convert a datetime value to
        the period index in model time.

        :param time_index: The datetime or integer index to validate
        :type time_index: Optional[Union[datetime, int]]
        :param time_list: A list of datetimes against which to validate time_index
        :type time_list: list
        :param start: The starting datetime in the out file (for diagnostics)
        :type start: datetime
        :param end: The ending datetime in the out file (for diagnostics)
        :type end: datetime
        :param report: The reporting interval in the out file (seconds, for diagnostics)
        :type report: int
        :param default_time: The default time_index to use if time_index is None (int or datetime)
        :type default_time: Union[datetime, int]
        :raises OutputException: If the time_index cannot be validated or is out-of-range
        :return: The integer index of the given time
        :rtype: int
        """

        max_index = len(time_list) - 1

        def _raise_out_of_range(arg):
            datetime_format = "%Y-%m-%d %H:%M:%S"
            msg = (
                f"{arg} does not exist in model output reporting time steps."
                f" The reporting time range is {start.strftime(datetime_format)} to "
                f"{end.strftime(datetime_format)} at increments of {report} seconds."
                f" Valid integer indices are 0..{max_index}."
            )
            raise OutputException(msg)

        resolved = time_index if time_index is not None else default_time

        if isinstance(resolved, datetime):
            if resolved in time_list:
                return time_list.index(resolved)
            _raise_out_of_range(resolved)

        if isinstance(resolved, int):
            if 0 <= resolved <= max_index:
                return resolved
            _raise_out_of_range(resolved)

        _raise_out_of_range(resolved)
        assert False, "unreachable"

    def open(self) -> bool:
        """
        Open a binary file

        :return: True if binary file was opened successfully
        :rtype: bool
        """
        if self.handle is None:
            self.handle = output.init()

        if not self.loaded:
            self.loaded = True
            output.open(self.handle, self.binfile)
            self.start = from_jd(output.get_start_date(self.handle) + 2415018.5)
            self.start = self.start.replace(microsecond=0)
            self.report = output.get_times(self.handle, shared_enum.Time.REPORT_STEP)
            self.period = output.get_times(self.handle, shared_enum.Time.NUM_PERIODS)
            self.end = self.start + timedelta(seconds=self.period * self.report)

        return True

    def close(self) -> bool:
        """
        Close an opened binary file

        :returns: True if binary file was closed successfully
        :rtype: bool
        """
        if self.handle or self.loaded:
            self.loaded = False
            self.delete_handle = True
            output.close(self.handle)

        return True

    # method used for context manager with statement
    def __enter__(self):
        self.open()
        return self

    # method used for context manager with statement
    def __exit__(self, *arg) -> NoReturn:
        self.close()

    @property
    def times(self) -> list:
        """
        Returns list of reporting timestep stored in model binary file

        :returns: list of datetime values for each reporting timestep
        :rtype: list

        Examples:

        >>> from pyswmm import Output
        >>>
        >>> with Output('tests/data/model_full_features.out') as out:
        ...     times = out.times
        ...     for time in times:
        ...         print(time)
        >>> 2015-11-01 14:01:00
        >>> 2015-11-01 14:02:00
        >>> 2015-11-01 14:03:00
        >>> 2015-11-01 14:04:00
        """
        if self._times is None:
            self._load_times()
        return self._times

    @output_open_handler
    def _load_times(self) -> NoReturn:
        """Load model reporting times into self._times"""
        self._times = list()
        for step in range(1, self.period + 1):
            self._times.append(self.start + timedelta(seconds=self.report) * step)

    @property
    def project_size(self) -> list:
        """
        Returns project size for model elements in the following order:
        [subcatchment, node, link, system, pollutant]

        :returns: list of numbers of each model type
                  [nSubcatchments, nNodes, nLinks, nSystems(1), nPollutants]
        :rtype: list

        Examples:

        >>> from pyswmm import Output
        >>>
        >>> with Output('tests/data/model_full_features.out') as out:
        ...     print(out.project_size)
        >>> [3, 4, 3, 1, 0]
        """
        if self._project_size is None:
            self._load_project_size()
        return self._project_size

    @output_open_handler
    def _load_project_size(self) -> NoReturn:
        """Load model size into self._project_size"""
        self._project_size = output.get_proj_size(self.handle)

    @property
    def subcatchments(self) -> dict:
        """
        Return a dict of subcatchments stored in SWMM output binary file

        :returns: dict of model subcatchment names with their indices as values
        :rtype: dict

        Examples:

        >>> from pyswmm import Output
        >>>
        >>> with Output('tests/data/model_full_features.out') as out:
        ...     print(out.subcatchments)
        >>> {'S1': 0, 'S2': 1, 'S3': 2}
        """
        if self._subcatchments is None:
            self._load_subcatchments()
        return self._subcatchments

    @output_open_handler
    def _load_subcatchments(self) -> NoReturn:
        """Load model size into self._project_size"""
        total = self.project_size[0]
        self._subcatchments = {
            self.object_name(shared_enum.ElementType.SUBCATCH, index): index
            for index in range(total)
        }

    @property
    def nodes(self) -> dict:
        """
        Return a dict of nodes stored in SWMM output binary file

        :returns: dict of model node names with their indices as values
        :rtype: dict

        Examples:

        >>> from pyswmm import Output
        >>>
        >>> with Output('tests/data/model_full_features.out') as out:
        ...     print(out.nodes)
        >>> {'J1': 0, 'J3': 1, 'J4': 2, 'J2': 3}
        """
        if self._nodes is None:
            self._load_nodes()
        return self._nodes

    @output_open_handler
    def _load_nodes(self) -> NoReturn:
        """Load model nodes into self._nodes"""
        total = self.project_size[1]
        self._nodes = {
            self.object_name(shared_enum.ElementType.NODE, index): index
            for index in range(total)
        }

    @property
    def links(self) -> dict:
        """Return a dict of links stored in SWMM output binary file

        :returns: dict of model link names with their indices as values
        :rtype: dict

        Examples:

        >>> from pyswmm import Output
        >>>
        >>> with Output('tests/data/model_full_features.out') as out:
        ...     print(out.links)
        >>> {'C1:C2': 0, 'C2': 1, 'C3': 2}
        """
        if self._links is None:
            self._load_links()
        return self._links

    @output_open_handler
    def _load_links(self) -> NoReturn:
        """Load model links into self._links"""
        total = self.project_size[2]
        self._links = {
            self.object_name(shared_enum.ElementType.LINK, index): index
            for index in range(total)
        }

    @property
    def pollutants(self) -> dict:
        """
        Return a dict of pollutants stored in SWMM output binary file

        :returns: dict of pollutant names with their indices as values
        :rtype: dict

        Examples:

        >>> from pyswmm import Output
        >>>
        >>> with Output('tests/data/model_full_features.out') as out:
        ...     print(out.pollutants)
        >>> {}
        """
        if self._pollutants is None:
            self._load_pollutants()
        return self._pollutants

    @output_open_handler
    def _load_pollutants(self) -> NoReturn:
        """Load model size into self._project_size"""
        total = self.project_size[4]
        self._pollutants = {
            self.object_name(shared_enum.ElementType.POLLUT, index): index
            for index in range(total)
        }

    @property
    @output_open_handler
    def units(self) -> int:
        """
        Return SWMM output binary file units for unit system, flow, and pollutants

        :returns: dict indicating units for unit system, flow, and pollutant
        :rtype: dict

        Examples:

        >>> from pyswmm import Output
        >>>
        >>> with Output('tests/data/model_full_features.out') as out:
        ...     print(out.units)
        >>> {'system': 'US', 'flow': 'CFS', 'pollutant': 'NONE'}
        """
        units_mapping = dict(
            system=shared_enum.UnitSystem,
            flow=shared_enum.FlowUnits,
            pollutant=shared_enum.ConcUnits,
        )
        return {
            attribute: units_mapping[attribute](unit).name
            for unit, attribute in zip(
                output.get_units(self.handle), units_mapping.keys()
            )
        }

    @property
    @output_open_handler
    def version(self) -> int:
        """
        Return SWMM version used to generate SWMM output binary file results

        :returns: integer representation of SWMM version used to make out file
        :rtype: int

        Examples:

        >>> from pyswmm import Output
        >>>
        >>> with Output('tests/data/model_full_features.out') as out:
        ...     print(out.unit)
        >>> 51013
        """
        return output.get_version(self.handle)

    @output_open_handler
    def object_name(self, object_type: int, index: int) -> str:
        """
        Get object name from SWMM output binary file using object index and object type

        :param object_type: object type from swmm.toolkit.shared_enum.ElementType:
                            SUBCATCH=0, NODE=1, LINK=2, SYSTEM=3, POLLUT=4
        :type object_type: int
        :param index: object index
        :type index: int
        :returns: object name
        :rtype: str

        Examples:

        >>> from swmm.toolkit.shared_enum import ElementType
        >>> from pyswmm import Output
        >>>
        >>> with Output('tests/data/model_full_features.out') as out:
        ...     print(out.object_name(ElementType.SUBCATCH, 0))
        ...     print(out.object_name(ElementType.NODE, 0))
        ...     print(out.object_name(ElementType.LINK, 0))
        >>> S1
        >>> J1
        >>> C1:C2
        """
        return output.get_elem_name(self.handle, object_type, index)

    @output_open_handler
    def subcatch_series(
        self,
        index: Union[int, str],
        attribute: shared_enum.SubcatchAttribute,
        start_index: Union[int, datetime, None] = None,
        end_index: Union[int, datetime, None] = None,
        end_exclusive: bool = False,
    ) -> dict:
        """
        Get subcatchment time series results for particular attribute. Specify series
        start index and end index to get desired time range.

        Note: Results are end-inclusive by default (end_index is included). Set end_exclusive=True to use Python-style [start, end) semantics and exclude the end.
              You can use pandas to convert the returned dict to a pandas Series object with datetimes as the index.

        :param index: subcatchment index or name
        :type index: Union[int, str]
        :param attribute: attribute from swmm.toolkit.shared_enum.SubcatchAttribute: RAINFALL, SNOW_DEPTH, EVAP_LOSS,
                          INFIL_LOSS, RUNOFF_RATE, GW_OUTFLOW_RATE, GW_TABLE_ELEV, SOIL_MOISTURE, POLLUT_CONC_0
        :type attribute: swmm.toolkit.shared_enum.SubcatchAttribute
        :param start_index: start datetime or index from which to return series, defaults to None
        :type start_index: Union[int, datetime, None], optional
        :param end_index: end datetime or index from which to return series, defaults to None
        :type end_index: Union[int, datetime, None], optional
        :param end_exclusive: If True, interpret end_index as an exclusive bound (Python slicing style), excluding the end timestep/index. Defaults to False (inclusive).
        :type end_exclusive: bool, optional
        :return: dict of attribute values with between start_index and end_index
                 with reporting timesteps as keys {datetime : value}
        :rtype: dict

        Examples:

        >>> from swmm.toolkit.shared_enum import SubcatchAttribute
        >>> from datetime import datetime
        >>> from pyswmm import Output
        >>>
        >>> with Output('tests/data/model_full_features.out') as out:
        ...     ts = out.subcatch_series('S1', SubcatchAttribute.RUNOFF_RATE, datetime(2015, 11, 1, 15), datetime(2015, 11, 1, 16))
        ...     for index in ts:
        ...         print(index, ts[index])
        >>> 2015-11-01 15:00:00 0.004305889364331961
        >>> 2015-11-01 15:01:00 0.004312679171562195
        >>> 2015-11-01 15:02:00 0.004319469444453716
        >>> 2015-11-01 15:03:00 0.00432625925168395
        """
        index = self.verify_index(index, self.subcatchments, "subcatchment")
        start_index = self.verify_time(
            start_index, self.times, self.start, self.end, self.report, 0
        )
        # Determine inclusive end index; default to last valid index
        if end_index is None:
            end_idx = self.verify_time(
                None, self.times, self.start, self.end, self.report, self.period - 1
            )
        elif end_exclusive:
            if isinstance(end_index, datetime):
                tmp_end = self.verify_time(
                    end_index,
                    self.times,
                    self.start,
                    self.end,
                    self.report,
                    self.period - 1,
                )
                end_idx = tmp_end - 1
            else:
                end_idx = end_index - 1
            if end_idx < start_index:
                return {}
            end_idx = self.verify_time(
                end_idx, self.times, self.start, self.end, self.report, self.period - 1
            )
        else:
            end_idx = self.verify_time(
                end_index,
                self.times,
                self.start,
                self.end,
                self.report,
                self.period - 1,
            )

        values = output.get_subcatch_series(
            self.handle, index, attribute, start_index, end_idx
        )
        return {
            time: value
            for time, value in zip(self.times[start_index : end_idx + 1], values)
        }

    @output_open_handler
    def node_series(
        self,
        index: Union[int, str],
        attribute: shared_enum.NodeAttribute,
        start_index: Union[int, datetime, None] = None,
        end_index: Union[int, datetime, None] = None,
        end_exclusive: bool = False,
    ) -> dict:
        """
        Get node time series results for particular attribute. Specify series
        start index and end index to get desired time range.

        Note: Results are end-inclusive by default (end_index is included). Set end_exclusive=True to use Python-style [start, end) semantics and exclude the end.
              You can use pandas to convert the returned dict to a pandas Series object with datetimes as the index.

        :param index: node index or name
        :type index: Union[int, str]
        :param attribute: attribute from swmm.toolkit.shared_enum.NodeAttribute: INVERT_DEPTH, HYDRAULIC_HEAD,
                          PONDED_VOLUME, LATERAL_INFLOW, TOTAL_INFLOW, FLOODING_LOSSES, POLLUT_CONC_0
        :type attribute: swmm.toolkit.shared_enum.NodeAttribute
        :param start_index: start datetime or index from which to return series, defaults to None
        :type start_index: Union[int, datetime, None], optional
        :param end_index: end datetime or index from which to return series, defaults to None
        :type end_index: Union[int, datetime, None], optional
        :param end_exclusive: If True, interpret end_index as an exclusive bound (Python slicing style), excluding the end timestep/index. Defaults to False (inclusive).
        :type end_exclusive: bool, optional
        :return: dict of attribute values with between start_index and end_index
                 with reporting timesteps as keys
        :rtype: dict {datetime : value}

        Examples:

        >>> from swmm.toolkit.shared_enum import NodeAttribute
        >>> from datetime import datetime
        >>> from pyswmm import Output
        >>>
        >>> with Output('tests/data/model_full_features.out') as out:
        ...     ts = out.node_series('J1', NodeAttribute.INVERT_DEPTH, datetime(2015, 11, 1, 15), datetime(2015, 11, 1, 16))
        ...     for index in ts:
        ...         print(index, ts[index])
        >>> 2015-11-01 15:00:00 15.0
        >>> 2015-11-01 15:01:00 15.0
        >>> 2015-11-01 15:02:00 15.0
        >>> 2015-11-01 15:03:00 15.0
        """

        index = self.verify_index(index, self.nodes, "node")
        start_index = self.verify_time(
            start_index, self.times, self.start, self.end, self.report, 0
        )
        # Determine inclusive end index; default to last valid index
        if end_index is None:
            end_idx = self.verify_time(
                None, self.times, self.start, self.end, self.report, self.period - 1
            )
        elif end_exclusive:
            if isinstance(end_index, datetime):
                tmp_end = self.verify_time(
                    end_index,
                    self.times,
                    self.start,
                    self.end,
                    self.report,
                    self.period - 1,
                )
                end_idx = tmp_end - 1
            else:
                end_idx = end_index - 1
            if end_idx < start_index:
                return {}
            end_idx = self.verify_time(
                end_idx, self.times, self.start, self.end, self.report, self.period - 1
            )
        else:
            end_idx = self.verify_time(
                end_index,
                self.times,
                self.start,
                self.end,
                self.report,
                self.period - 1,
            )

        values = output.get_node_series(
            self.handle, index, attribute, start_index, end_idx
        )
        return {
            time: value
            for time, value in zip(self.times[start_index : end_idx + 1], values)
        }

    @output_open_handler
    def link_series(
        self,
        index: Union[int, str],
        attribute: shared_enum.LinkAttribute,
        start_index: Union[int, datetime, None] = None,
        end_index: Union[int, datetime, None] = None,
        end_exclusive: bool = False,
    ) -> dict:
        """
        Get link time series results for particular attribute. Specify series
        start index and end index to get desired time range.

        Note: Results are end-inclusive by default (end_index is included). Set end_exclusive=True to use Python-style [start, end) semantics and exclude the end.
              You can use pandas to convert the returned dict to a pandas Series object with datetimes as the index.

        :param index: link index or name
        :type index: Union[int, str]
        :param attribute: attribute from swmm.toolkit.shared_enum.LinkAttribute: FLOW_RATE, FLOW_DEPTH,
                          FLOW_VELOCITY, FLOW_VOLUME, CAPACITY, POLLUT_CONC_0
        :type attribute: swmm.toolkit.shared_enum.LinkAttribute
        :param start_index: start datetime or index from which to return series, defaults to None
        :type start_index: Union[int, datetime, None], optional
        :param end_index: end datetime or index from which to return series, defaults to None
        :type end_index: Union[int, datetime, None], optional
        :param end_exclusive: If True, interpret end_index as an exclusive bound (Python slicing style), excluding the end timestep/index. Defaults to False (inclusive).
        :type end_exclusive: bool, optional
        :return: dict of attribute values with between start_index and end_index
                 with reporting timesteps as keys
        :rtype: dict {datetime : value}

        Examples:

        >>> from swmm.toolkit.shared_enum import LinkAttribute
        >>> from datetime import datetime
        >>> from pyswmm import Output
        >>>
        >>> with Output('tests/data/model_full_features.out') as out:
        ...     ts = out.link_series('C2', LinkAttribute.FLOW_RATE, datetime(2015, 11, 1, 15), datetime(2015, 11, 1, 16))
        ...     for index in ts:
        ...         print(index, ts[index])
        >>> 2015-11-01 15:00:00 8.226319313049316
        >>> 2015-11-01 15:01:00 8.226363182067871
        >>> 2015-11-01 15:02:00 8.226407051086426
        >>> 2015-11-01 15:03:00 8.22645092010498
        """
        index = self.verify_index(index, self.links, "link")
        start_index = self.verify_time(
            start_index, self.times, self.start, self.end, self.report, 0
        )
        # Determine inclusive end index; default to last valid index
        if end_index is None:
            end_idx = self.verify_time(
                None, self.times, self.start, self.end, self.report, self.period - 1
            )
        elif end_exclusive:
            if isinstance(end_index, datetime):
                tmp_end = self.verify_time(
                    end_index,
                    self.times,
                    self.start,
                    self.end,
                    self.report,
                    self.period - 1,
                )
                end_idx = tmp_end - 1
            else:
                end_idx = end_index - 1
            if end_idx < start_index:
                return {}
            end_idx = self.verify_time(
                end_idx, self.times, self.start, self.end, self.report, self.period - 1
            )
        else:
            end_idx = self.verify_time(
                end_index,
                self.times,
                self.start,
                self.end,
                self.report,
                self.period - 1,
            )

        values = output.get_link_series(
            self.handle, index, attribute, start_index, end_idx
        )
        return {
            time: value
            for time, value in zip(self.times[start_index : end_idx + 1], values)
        }

    @output_open_handler
    def system_series(
        self,
        attribute: shared_enum.SystemAttribute,
        start_index: Union[int, datetime, None] = None,
        end_index: Union[int, datetime, None] = None,
        end_exclusive: bool = False,
    ) -> dict:
        """
        Get system time series results for particular attribute. Specify series
        start index and end index to get desired time range.

        Note: Results are end-inclusive by default (end_index is included). Set end_exclusive=True to use Python-style [start, end) semantics and exclude the end.
              You can use pandas to convert the returned dict to a pandas Series object with datetimes as the index.

        :param attribute: attribute from swmm.toolkit.shared_enum.SystemAttribute: AIR_TEMP, RAINFALL, SNOW_DEPTH,
                          EVAP_INFIL_LOSS, RUNOFF_FLOW, DRY_WEATHER_INFLOW, GW_INFLOW, RDII_INFLOW, DIRECT_INFLOW,
                          TOTAL_LATERAL_INFLOW, FLOOD_LOSSES, OUTFALL_FLOWS, VOLUME_STORED, EVAP_RATE
        :type attribute: swmm.toolkit.shared_enum.SystemAttribute
        :param start_index: start datetime or index from which to return series, defaults to None
        :type start_index: Union[int, datetime, None], optional
        :param end_index: end datetime or index from which to return series, defaults to None
        :type end_index: Union[int, datetime, None], optional
        :param end_exclusive: If True, interpret end_index as an exclusive bound (Python slicing style), excluding the end timestep/index. Defaults to False (inclusive).
        :type end_exclusive: bool, optional
        :return: dict of attribute values with between start_index and end_index
                 with reporting timesteps as keys
        :rtype: dict {datetime : value}

        Examples:

        >>> from swmm.toolkit.shared_enum import SystemAttribute
        >>> from datetime import datetime
        >>> from pyswmm import Output
        >>>
        >>> with Output('tests/data/model_full_features.out') as out:
        ...     ts = out.system_series(SystemAttribute.RUNOFF_FLOW, datetime(2015, 11, 1, 15), datetime(2015, 11, 1, 16))
        ...     for index in ts:
        ...         print(index, ts[index])
        >>> 2015-11-01 15:00:00 0.02255748212337494
        >>> 2015-11-01 15:01:00 0.022702988237142563
        >>> 2015-11-01 15:02:00 0.022848498076200485
        >>> 2015-11-01 15:03:00 0.022994007915258408
        """
        start_index = self.verify_time(
            start_index, self.times, self.start, self.end, self.report, 0
        )
        # Determine inclusive end index; default to last valid index
        if end_index is None:
            end_idx = self.verify_time(
                None, self.times, self.start, self.end, self.report, self.period - 1
            )
        elif end_exclusive:
            if isinstance(end_index, datetime):
                tmp_end = self.verify_time(
                    end_index,
                    self.times,
                    self.start,
                    self.end,
                    self.report,
                    self.period - 1,
                )
                end_idx = tmp_end - 1
            else:
                end_idx = end_index - 1
            if end_idx < start_index:
                return {}
            end_idx = self.verify_time(
                end_idx, self.times, self.start, self.end, self.report, self.period - 1
            )
        else:
            end_idx = self.verify_time(
                end_index,
                self.times,
                self.start,
                self.end,
                self.report,
                self.period - 1,
            )

        values = output.get_system_series(self.handle, attribute, start_index, end_idx)
        return {
            time: value
            for time, value in zip(self.times[start_index : end_idx + 1], values)
        }

    @output_open_handler
    def subcatch_attribute(
        self,
        attribute: shared_enum.SubcatchAttribute,
        time_index: Union[int, datetime, None] = None,
    ) -> dict:
        """
        For all subcatchments at given time, get a particular attribute.

        :param attribute: attribute from swmm.toolkit.shared_enum.SubcatchAttribute: RAINFALL, SNOW_DEPTH, EVAP_LOSS,
                          INFIL_LOSS, RUNOFF_RATE, GW_OUTFLOW_RATE, GW_TABLE_ELEV, SOIL_MOISTURE, POLLUT_CONC_0
        :type attribute: swmm.toolkit.shared_enum.SubcatchAttribute
        :param time_index: datetime or simulation index, defaults to None
        :type time_index: Union[int, datetime, None]
        :returns: dict of attribute value for all subcatchments at given timestep
        :rtype: dict {subcatchment: value}

        Examples:

        >>> from swmm.toolkit.shared_enum import SubcatchAttribute
        >>> from datetime import datetime
        >>> from pyswmm import Output
        >>>
        >>> with Output('tests/data/model_full_features.out') as out:
        ...     data = out.subcatch_attribute(SubcatchAttribute.RUNOFF_RATE, datetime(2015, 11, 1, 16))
        ...     for object in data:
        ...         print(object, data[object])
        >>> S1 0.004410634282976389
        >>> S2 0.008775550872087479
        >>> S3 0.012964698486030102
        """

        time_index = self.verify_time(
            time_index, self.times, self.start, self.end, self.report, 0
        )

        values = output.get_subcatch_attribute(self.handle, time_index, attribute)
        return {sub: value for sub, value in zip(self.subcatchments, values)}

    @output_open_handler
    def node_attribute(
        self,
        attribute: shared_enum.NodeAttribute,
        time_index: Union[int, datetime, None] = None,
    ) -> dict:
        """
        For all nodes at given time, get a particular attribute.

        :param attribute: attribute from swmm.toolkit.shared_enum.NodeAttribute: INVERT_DEPTH, HYDRAULIC_HEAD,
                          PONDED_VOLUME, LATERAL_INFLOW, TOTAL_INFLOW, FLOODING_LOSSES, POLLUT_CONC_0
        :type attribute: swmm.toolkit.shared_enum.NodeAttribute
        :param time_index: datetime or simulation index, defaults to None
        :type time_index: Union[int, datetime, None]
        :returns: dict of attribute values for all nodes at given timestep
        :rtype: dict {node:value}

        Examples:

        >>> from swmm.toolkit.shared_enum import NodeAttribute
        >>> from datetime import datetime
        >>> from pyswmm import Output
        >>>
        >>> with Output('tests/data/model_full_features.out') as out:
        ...     data = out.node_attribute(NodeAttribute.INVERT_DEPTH, datetime(2015, 11, 1, 16))
        ...     for object in data:
        ...         print(object, data[object])
        >>> J1 15.0
        >>> J3 1.9746614694595337
        >>> J4 0.0
        >>> J2 0.0009783204877749085
        """

        time_index = self.verify_time(
            time_index, self.times, self.start, self.end, self.report, 0
        )

        values = output.get_node_attribute(self.handle, time_index, attribute)
        return {node: value for node, value in zip(self.nodes, values)}

    @output_open_handler
    def link_attribute(
        self,
        attribute: shared_enum.LinkAttribute,
        time_index: Union[int, datetime, None] = None,
    ):
        """
        For all links at given time, get a particular attribute.

        :param attribute: attribute from swmm.toolkit.shared_enum.LinkAttribute: FLOW_RATE, FLOW_DEPTH,
                          FLOW_VELOCITY, FLOW_VOLUME, CAPACITY, POLLUT_CONC_0
        :type attribute: swmm.toolkit.shared_enum.LinkAttribute
        :param time_index: datetime or simulation index, defaults to None
        :type time_index: Union[int, datetime, None]
        :returns: dict of attribute values for all nodes at given timestep
        :rtype: dict {link : value}

        Examples:

        >>> from swmm.toolkit.shared_enum import LinkAttribute
        >>> from datetime import datetime
        >>> from pyswmm import Output
        >>>
        >>> with Output('tests/data/model_full_features.out') as out:
        ...     data = out.link_attribute(LinkAttribute.FLOW_RATE, datetime(2015, 11, 1, 16))
        ...     for object in data:
        ...         print(object, data[object])
        >>> C1:C2 7.218499660491943
        >>> C2 8.227274894714355
        >>> C3 9.240239143371582
        """

        time_index = self.verify_time(
            time_index, self.times, self.start, self.end, self.report, 0
        )

        values = output.get_link_attribute(self.handle, time_index, attribute)
        return {link: value for link, value in zip(self.links, values)}

    # @output_open_handler
    # def system_attribute(
    #     self,
    #     attribute: shared_enum.SystemAttribute,
    #     time_index: Union[int, datetime, None] = None,
    # ):
    #     """
    #     At given time, get a particular system attribute.
    #
    #     :param attribute: attribute from swmm.toolkit.shared_enum.SystemAttribute: AIR_TEMP, RAINFALL,
    #                       SNOW_DEPTH, EVAP_INFIL_LOSS, RUNOFF_FLOW, DRY_WEATHER_INFLOW, GW_INFLOW, RDII_INFLOW,
    #                       DIRECT_INFLOW, TOTAL_LATERAL_INFLOW, FLOOD_LOSSES, OUTFALL_FLOWS, VOLUME_STORED, EVAP_RATE
    #     :type attribute: swmm.toolkit.shared_enum.SystemAttribute
    #     :param time_index: datetime or simulation index, defaults to None
    #     :type time_index: Union[int, datetime, None]
    #     :returns: dict of attribute value for system at given timestep
    #     :rtype: dict of {"system",value}
    #
    #     Not usable due to error in swmm toolkit
    #     """
    #
    #     time_index = self.verify_time(
    #         time_index, self.times, self.start, self.end, self.report, 0
    #     )
    #
    #     value = output.get_system_attribute(self.handle, time_index, attribute)
    #     return {"system": value}

    @output_open_handler
    def subcatch_result(
        self, index: Union[int, str], time_index: Union[int, datetime, None] = None
    ):
        """
        For a subcatchment at given time, get all attributes.

        :param index: subcatchment name or index
        :type index: Union[int, str]
        :param time_index: datetime or simulation index, defaults to None
        :type time_index: Union[int, datetime, None]
        :returns: dict of attributes for a subcatchment at given timestep
        :rtype: dict {attribute:value}

        Examples:

        >>> from pyswmm import Output
        >>> from datetime import datetime
        >>>
        >>> with Output('tests/data/model_full_features.out') as out:
        ...     data = out.subcatch_result('S1', datetime(2015, 11, 1, 15))
        ...     for attribute in data:
        ...         print(attribute, data[attribute])
        >>> SubcatchAttribute.RAINFALL 0.017500000074505806
        >>> SubcatchAttribute.SNOW_DEPTH 0.0
        >>> SubcatchAttribute.EVAP_LOSS 0.0
        >>> SubcatchAttribute.INFIL_LOSS 0.0
        >>> SubcatchAttribute.RUNOFF_RATE 0.004305889364331961
        >>> SubcatchAttribute.GW_OUTFLOW_RATE 0.0
        >>> SubcatchAttribute.GW_TABLE_ELEV 0.0
        >>> SubcatchAttribute.SOIL_MOISTURE 0.0
        """
        index = self.verify_index(index, self.subcatchments, "subcatchment")
        time_index = self.verify_time(
            time_index, self.times, self.start, self.end, self.report, 0
        )

        values = output.get_subcatch_result(self.handle, time_index, index)
        return {
            attr: value for attr, value in zip(shared_enum.SubcatchAttribute, values)
        }

    @output_open_handler
    def node_result(
        self, index: Union[int, str], time_index: Union[int, datetime, None] = None
    ):
        """
        For a node at given time, get all attributes.

        :param index: node name or index
        :type index: Union[int, str]
        :param time_index: datetime or simulation index, defaults to None
        :type time_index: Union[int, datetime, None]
        :returns: dict of attributes for a node at given timestep
        :rtype: dict {attribute:value}

        Examples:

        >>> from pyswmm import Output
        >>> from datetime import datetime
        >>>
        >>> with Output('tests/data/model_full_features.out') as out:
        ...     data = out.node_result('J1', datetime(2015, 11, 1, 15))
        ...     for attribute in data:
        ...         print(attribute, data[attribute])
        >>> NodeAttribute.INVERT_DEPTH 15.0
        >>> NodeAttribute.HYDRAULIC_HEAD 35.72800064086914
        >>> NodeAttribute.PONDED_VOLUME 0.0
        >>> NodeAttribute.LATERAL_INFLOW 9.004305839538574
        >>> NodeAttribute.TOTAL_INFLOW 9.004305839538574
        >>> NodeAttribute.FLOODING_LOSSES 1.7858062982559204
        """
        index = self.verify_index(index, self.nodes, "node")
        time_index = self.verify_time(
            time_index, self.times, self.start, self.end, self.report, 0
        )

        values = output.get_node_result(self.handle, time_index, index)
        return {attr: value for attr, value in zip(shared_enum.NodeAttribute, values)}

    @output_open_handler
    def link_result(
        self, index: Union[int, str], time_index: Union[int, datetime, None] = None
    ):
        """
        For a link at given time, get all attributes.

        :param index: link name or index
        :type index: Union[int, str]
        :param time_index: datetime or simulation index, defaults to None
        :type time_index: Union[int, datetime, None]
        :returns: dict of attributes for a link at given timestep
        :rtype: dict {attribute:value}

        Examples:

        >>> from pyswmm import Output
        >>> from datetime import datetime
        >>>
        >>> with Output('tests/data/model_full_features.out') as out:
        ...     data = out.link_result('C2', datetime(2015, 11, 1, 15))
        ...     for attribute in data:
        ...         print(attribute, data[attribute])
        >>> LinkAttribute.FLOW_RATE 8.226319313049316
        >>> LinkAttribute.FLOW_DEPTH 0.0
        >>> LinkAttribute.FLOW_VELOCITY 0.0
        >>> LinkAttribute.FLOW_VOLUME 0.0
        >>> LinkAttribute.CAPACITY 1.0
        """
        index = self.verify_index(index, self.links, "link")
        time_index = self.verify_time(
            time_index, self.times, self.start, self.end, self.report, 0
        )

        values = output.get_link_result(self.handle, time_index, index)
        return {attr: value for attr, value in zip(shared_enum.LinkAttribute, values)}

    @output_open_handler
    def system_result(self, time_index: Union[int, datetime, None] = None):
        """
        At a given time, get all system attributes.

        :param time_index: datetime or simulation index, defaults to None
        :type time_index: Union[int, datetime, None]
        :returns: dict of attributes for the system at given timestep
        :rtype: dict {attribute:value}

        Examples:

        >>> from pyswmm import Output
        >>> from datetime import datetime
        >>>
        >>> with Output('tests/data/model_full_features.out') as out:
        ...     data = out.system_result(datetime(2015, 11, 1, 15))
        ...     for attribute in data:
        ...         print(attribute, data[attribute])
        >>> SystemAttribute.AIR_TEMP 70.0
        >>> SystemAttribute.RAINFALL 0.017500001937150955
        >>> SystemAttribute.SNOW_DEPTH 0.0
        >>> SystemAttribute.EVAP_INFIL_LOSS 0.0
        >>> SystemAttribute.RUNOFF_FLOW 0.02255748212337494
        >>> SystemAttribute.DRY_WEATHER_INFLOW 8.0
        >>> SystemAttribute.GW_INFLOW 0.0
        >>> SystemAttribute.RDII_INFLOW 0.0
        >>> SystemAttribute.DIRECT_INFLOW 3.0
        >>> SystemAttribute.TOTAL_LATERAL_INFLOW 11.022557258605957
        >>> SystemAttribute.FLOOD_LOSSES 1.7858062982559204
        >>> SystemAttribute.OUTFALL_FLOWS 9.23674488067627
        >>> SystemAttribute.VOLUME_STORED 97.05753326416016
        >>> SystemAttribute.EVAP_RATE 0.0
        """
        dummy_index = 0
        time_index = self.verify_time(
            time_index, self.times, self.start, self.end, self.report, 0
        )

        values = output.get_system_result(self.handle, time_index, dummy_index)
        return {attr: value for attr, value in zip(shared_enum.SystemAttribute, values)}


class OutAttributeBase:
    """Baseclass for SubcatchSeries, NodeSeries, LinkSeries."""

    def __init__(self, out_handle: Output):
        if not isinstance(out_handle, Output):
            raise (TypeError("Invalid Outfile Handle"))
        self._handle = out_handle
        self._attr_group = None

    def __dir__(self):
        return super().__dir__() + [val.name.lower() for val in self._attr_group]

    def __getattr__(self, attr) -> dict[datetime.datetime, float]:
        if attr.upper() not in [item.name for item in self._attr_group]:
            raise (AttributeError("Invalid Property: {}".format(attr)))
        else:
            attr_select = getattr(self._attr_group, attr.upper())
            ts = self._series_type(attr_select)
        return ts


class SubcatchSeries(OutAttributeBase):
    """
    Get a subcatchment time series.  New to PySWMM-v2!

    Note: you can use pandas to convert dict to a pandas Series object with dict keys as index

    :return: dict of attribute values with between start_index and end_index
             with reporting timesteps as keys
    :rtype: dict {datetime : value}

    Examples:

    .. code-block:: python

        from pyswmm import Output, SubcatchSeries

        with Output('tests/data/model_full_features.out') as out:
            ts1 = SubcatchSeries(out)['S1'].rainfall
            ts2 = SubcatchSeries(out)['S1'].snow_depth
            ts3 = SubcatchSeries(out)['S1'].evap_loss
            ts4 = SubcatchSeries(out)['S1'].infil_loss
            ts5 = SubcatchSeries(out)['S1'].runoff_rate
            ts6 = SubcatchSeries(out)['S1'].gw_outflow_rate
            ts7 = SubcatchSeries(out)['S1'].gw_table_elev
            ts8 = SubcatchSeries(out)['S1'].soil_moisture
            ts9 = SubcatchSeries(out)['S1'].pollut_conc_0
    """

    rainfall: dict[datetime.datetime, float]
    snow_depth: dict[datetime.datetime, float]
    evap_loss: dict[datetime.datetime, float]
    infil_loss: dict[datetime.datetime, float]
    runoff_rate: dict[datetime.datetime, float]
    gw_outflow_rate: dict[datetime.datetime, float]
    gw_table_elev: dict[datetime.datetime, float]
    soil_moisture: dict[datetime.datetime, float]
    pollut_conc_0: dict[datetime.datetime, float]

    def __init__(self, out_handle):
        super().__init__(out_handle)
        self._attr_group = shared_enum.SubcatchAttribute
        self._idname = None

    def __dir__(self):
        return super().__dir__() + [val.name.lower() for val in self._attr_group]

    def __getitem__(self, idname):
        self._idname = idname
        return self

    def _series_type(self, attr_select):
        return self._handle.subcatch_series(self._idname, attr_select)


class NodeSeries(OutAttributeBase):
    """
    Get a node time series.  New to PySWMM-v2!

    Note: you can use pandas to convert dict to a pandas Series object with dict keys as index

    :return: dict of attribute values with between start_index and end_index
             with reporting timesteps as keys
    :rtype: dict {datetime : value}

    Examples:

    .. code-block:: python

        from pyswmm import Output, NodeSeries

        with Output('tests/data/model_full_features.out') as out:
            ts1 = NodeSeries(out)['J1'].invert_depth
            ts2 = NodeSeries(out)['J1'].hydraulic_head
            ts3 = NodeSeries(out)['J1'].ponded_volume
            ts4 = NodeSeries(out)['J1'].lateral_inflow
            ts5 = NodeSeries(out)['J1'].total_inflow
            ts6 = NodeSeries(out)['J1'].flooding_losses
            ts7 = NodeSeries(out)['J1'].pollut_conc_0
    """

    invert_depth: dict[datetime.datetime, float]
    hydraulic_head: dict[datetime.datetime, float]
    ponded_volume: dict[datetime.datetime, float]
    lateral_inflow: dict[datetime.datetime, float]
    total_inflow: dict[datetime.datetime, float]
    flooding_losses: dict[datetime.datetime, float]
    pollut_conc_0: dict[datetime.datetime, float]

    def __init__(self, out_handle):
        super().__init__(out_handle)
        self._attr_group = shared_enum.NodeAttribute
        self._idname = None

    def __dir__(self):
        return super().__dir__() + [val.name.lower() for val in self._attr_group]

    def __getitem__(self, idname):
        self._idname = idname
        return self

    def _series_type(self, attr_select):
        return self._handle.node_series(self._idname, attr_select)


class LinkSeries(OutAttributeBase):
    """
    Get a link time series.  New to PySWMM-v2!

    Note: you can use pandas to convert dict to a pandas Series object with dict keys as index

    :return: dict of attribute values with between start_index and end_index
             with reporting timesteps as keys
    :rtype: dict {datetime : value}

    Examples:

    .. code-block:: python

        from pyswmm import Output, LinkSeries

        with Output('tests/data/model_full_features.out') as out:
            ts1 = LinkSeries(out)['C1:C2'].flow_rate
            ts2 = LinkSeries(out)['C1:C2'].flow_depth
            ts3 = LinkSeries(out)['C1:C2'].flow_velocity
            ts4 = LinkSeries(out)['C1:C2'].flow_volume
            ts5 = LinkSeries(out)['C1:C2'].capacity
            ts6 = LinkSeries(out)['C1:C2'].pollut_conc_0

    """

    flow_rate: dict[datetime.datetime, float]
    flow_depth: dict[datetime.datetime, float]
    flow_velocity: dict[datetime.datetime, float]
    flow_volume: dict[datetime.datetime, float]
    capacity: dict[datetime.datetime, float]
    pollut_conc_0: dict[datetime.datetime, float]

    def __init__(self, out_handle):
        super().__init__(out_handle)
        self._attr_group = shared_enum.LinkAttribute
        self._idname = None

    def __dir__(self):
        return super().__dir__() + [val.name.lower() for val in self._attr_group]

    def __getitem__(self, idname):
        self._idname = idname
        return self

    def _series_type(self, attr_select):
        return self._handle.link_series(self._idname, attr_select)


class SystemSeries(OutAttributeBase):
    """
    Get a system time series.  New to PySWMM-v2!

    Note: you can use pandas to convert dict to a pandas Series object with dict keys as index

    :return: dict of attribute values with between start_index and end_index
             with reporting timesteps as keys
    :rtype: dict {datetime : value}

    Examples:

    .. code-block:: python

        from pyswmm import Output, SystemSeries

        with Output('tests/data/model_full_features.out') as out:
            ts1 = SystemSeries(out).air_temp
            ts2 = SystemSeries(out).rainfall
            ts3 = SystemSeries(out).snow_depth
            ts4 = SystemSeries(out).evap_infil_loss
            ts5 = SystemSeries(out).runoff_flow
            ts6 = SystemSeries(out).dry_weather_inflow
            ts7 = SystemSeries(out).gw_inflow
            ts8 = SystemSeries(out).rdii_inflow
            ts9 = SystemSeries(out).direct_inflow
            ts10 = SystemSeries(out).total_lateral_inflow
            ts11 = SystemSeries(out).flood_losses
            ts12 = SystemSeries(out).outfall_flows
            ts13 = SystemSeries(out).volume_stored
            ts14 = SystemSeries(out).evap_rate
            ts15 = SystemSeries(out).ptnl_evap_rate
    """

    air_temp: dict[datetime.datetime, float]
    rainfall: dict[datetime.datetime, float]
    snow_depth: dict[datetime.datetime, float]
    evap_infil_loss: dict[datetime.datetime, float]
    runoff_flow: dict[datetime.datetime, float]
    dry_weather_inflow: dict[datetime.datetime, float]
    gw_inflow: dict[datetime.datetime, float]
    rdii_inflow: dict[datetime.datetime, float]
    direct_inflow: dict[datetime.datetime, float]
    total_lateral_inflow: dict[datetime.datetime, float]
    flood_losses: dict[datetime.datetime, float]
    outfall_flows: dict[datetime.datetime, float]
    volume_stored: dict[datetime.datetime, float]
    evap_rate: dict[datetime.datetime, float]
    ptnl_evap_rate: dict[datetime.datetime, float]

    def __init__(self, out_handle):
        super().__init__(out_handle)
        self._attr_group = shared_enum.SystemAttribute

    def __dir__(self):
        return super().__dir__() + [val.name.lower() for val in self._attr_group]

    def _series_type(self, attr_select):
        return self._handle.system_series(attr_select)



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2023 Bryant E. McDonnell
#
# Licensed under the terms of the BSD2 License
# See LICENSE.txt for details
# -----------------------------------------------------------------------------
"""Python setup.py installer script."""

# Standard library imports
import ast
import os
import sys

# Third party imports
from setuptools import find_packages, setup

HERE = os.path.abspath(os.path.dirname(__file__))


def get_version(module="pyswmm"):
    """Get version."""
    with open(os.path.join(HERE, module, "__init__.py"), "r") as f:
        data = f.read()
    lines = data.split("\n")
    for line in lines:
        if line.startswith("VERSION_INFO"):
            version_tuple = ast.literal_eval(line.split("=")[-1].strip())
            version = ".".join(map(str, version_tuple))
            break
    return version


def get_description():
    """Get long description."""
    with open(os.path.join(HERE, "README.md"), "r") as f:
        data = f.read()
    return data


def read_requirements(filename="requirements.txt"):
    with open(os.path.join(os.path.dirname(__file__), filename)) as f:
        lines = f.read().splitlines()
    # Strip comments and blank lines
    return [line.strip() for line in lines if line.strip() and not line.startswith("#")]


setup(
    name="pyswmm",
    version=get_version(),
    description="Python Wrapper for SWMM5 API",
    long_description=get_description(),
    long_description_content_type="text/markdown",
    url="https://www.pyswmm.org",
    author="Bryant E. McDonnell (See AUTHORS)",
    install_requires=read_requirements(),
    extras_require={
        "swmm5.1.14": ["swmm-toolkit==0.9.1"],
        "swmm5.1.15": ["swmm-toolkit==0.10.0"],
        "swmm5.2.0": ["swmm-toolkit==0.11.0"],
        "swmm5.2.1": ["swmm-toolkit==0.12.0"],
        "swmm5.2.2": ["swmm-toolkit==0.13.0"],
        "swmm5.2.3": ["swmm-toolkit==0.14.0"],
        "swmm5.2.4": ["swmm-toolkit<0.17.0"],
    },
    packages=find_packages(exclude=["contrib", "docs"]),
    package_data={"": ["LICENSE.txt", "AUTHORS", "tests/data/*.inp", "tests/*.py"]},
    include_package_data=True,
    license="BSD2 License",
    keywords="swmm5, swmm, hydraulics, hydrology, modeling, collection system",
    classifiers=[
        "Topic :: Scientific/Engineering",
        "Topic :: Documentation :: Sphinx",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: Unix",
        "Operating System :: MacOS",
        "License :: OSI Approved :: BSD License",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Programming Language :: Python",
        "Development Status :: 5 - Production/Stable",
    ],
)



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2014 Bryant E. McDonnell
#
# Licensed under the terms of the BSD2 License
# See LICENSE.txt for details
# -----------------------------------------------------------------------------
"""SWMM ouput toolkit api."""

# Standard library imports
from datetime import datetime, timedelta
import ctypes
import os

# Local imports
import pyswmm.toolkitapi as tka


class OutReaderNotImplementedYet(Exception):
    """ """

    def __init__(self, error_message):
        self.message = error_message

    def __str__(self):
        return self.error_message


class _Opaque(ctypes.Structure):
    """Used soley for passing the pointer to the smoapu struct to API."""

    pass


class SWMMBinReader(object):
    """Instantiate python Wrapper Object and build Wrapper functions."""

    raise OutReaderNotImplementedYet("Output Reader has not been implemented")

    def __init__(self):
        """Instantiate python Wrapper Object and build Wrapper functions."""

        def get_pkgpath():
            import toolkitapi as tkp

            return os.path.dirname(tkp.__file__.replace("\\", "/"))

        try:
            # Later Check for OS Type
            dllname = "outputAPI_winx86.dll"
            # when platform detection is enabled, dllname can be changed
            dllLoc = get_pkgpath() + "/lib/windows/" + dllname
            self.swmmdll = ctypes.CDLL(dllLoc)
        except Exception:
            raise (Exception("Failed to Open Linked Library"))

        # Initializing DLL Function List
        # Initialize Pointer to smoapi
        self._initsmoapi = self.swmmdll.SMO_init
        self._initsmoapi.restype = ctypes.POINTER(_Opaque)

        # Open File Function Handle
        self._openBinFile = self.swmmdll.SMO_open
        self._free = self.swmmdll.SMO_free
        self._close = self.swmmdll.SMO_close

        # Project Data
        self._getProjectSize = self.swmmdll.SMO_getProjectSize
        self._getTimes = self.swmmdll.SMO_getTimes
        self._getStartTime = self.swmmdll.SMO_getStartTime
        self._getUnits = self.swmmdll.SMO_getUnits

        # Object ID Function Handles
        self._getIDs = self.swmmdll.SMO_getElementName

        # Object Series Function Handles
        self._getSubcatchSeries = self.swmmdll.SMO_getSubcatchSeries
        self._getNodeSeries = self.swmmdll.SMO_getNodeSeries
        self._getLinkSeries = self.swmmdll.SMO_getLinkSeries
        self._getSystemSeries = self.swmmdll.SMO_getSystemSeries

        # Object Attribure Function Handles
        self._getSubcatchAttribute = self.swmmdll.SMO_getSubcatchAttribute
        self._getNodeAttribute = self.swmmdll.SMO_getNodeAttribute
        self._getLinkAttribute = self.swmmdll.SMO_getLinkAttribute
        self._getSystemAttribute = self.swmmdll.SMO_getSystemAttribute

        # Object Result Function Handles
        self._getSubcatchResult = self.swmmdll.SMO_getSubcatchResult
        self._getNodeResult = self.swmmdll.SMO_getNodeResult
        self._getLinkResult = self.swmmdll.SMO_getLinkResult
        self._getSystemResult = self.swmmdll.SMO_getSystemResult

        # Array Builder
        self._newOutValueArray = self.swmmdll.SMO_newOutValueArray
        self._newOutValueArray.argtypes = [
            ctypes.POINTER(_Opaque),
            ctypes.c_int,
            ctypes.c_int,
            ctypes.POINTER(ctypes.c_int),
            ctypes.POINTER(ctypes.c_int),
        ]
        self._newOutValueArray.restype = ctypes.POINTER(ctypes.c_float)

        # Series Builder
        self._newOutValueSeries = self.swmmdll.SMO_newOutValueSeries
        self._newOutValueSeries.argtypes = [
            ctypes.POINTER(_Opaque),
            ctypes.c_int,
            ctypes.c_int,
            ctypes.POINTER(ctypes.c_int),
            ctypes.POINTER(ctypes.c_int),
        ]
        self._newOutValueSeries.restype = ctypes.POINTER(ctypes.c_float)

        # SWMM Date num 2 String
        self.SWMMdateToStr = self.swmmdll.datetime_dateToStr

        # SWMM Time num 2 String
        self.SWMMtimeToStr = self.swmmdll.datetime_timeToStr

    def OpenBinFile(self, OutLoc):
        """
        Opens SWMM5 binary output file.

        :param str OutLoc: Path to Binary output file

        :return: None

        Examples:

        >>> OutputFile = SWMMBinReader()
        >>> OutputFile.OpenBinFile("outputfile.out")
        """
        self.smoapi = self._initsmoapi()
        ErrNo = self._openBinFile(self.smoapi, OutLoc)
        if ErrNo != 0:
            error_msg = "API ErrNo {0}:{1}".format(ErrNo, tka.DLLErrorKeys[ErrNo])
            raise Exception(error_msg)

    def CloseBinFile(self):
        """
        Closes binary output file and cleans up member variables.

        :returns: None

        Examples:

        >>> OutputFile = SWMMBinReader()
        >>> OutputFile.OpenBinFile("outputfile.out")
        >>> OutputFile.CloseBinFile()
        """
        ErrNo = self._close(self.smoapi)

        if hasattr(self, "SubcatchmentIDs"):
            delattr(self, "SubcatchmentIDs")
        if hasattr(self, "NodeIDs"):
            delattr(self, "NodeIDs")
        if hasattr(self, "LinkIDs"):
            delattr(self, "LinkIDs")
        if hasattr(self, "PollutantIDs"):
            delattr(self, "PollutantIDs")

        if ErrNo != 0:
            error_msg = "API ErrNo {0}:{1}".format(
                ErrNo.value, tka.DLLErrorKeys[ErrNo.value]
            )
            raise Exception(error_msg)

    def _get_SubcatchIDs(self):
        """Generates member Element IDs dictionary for Subcatchments."""
        self.SubcatchmentIDs = {}
        subcatchment_count = tka.SMO_elementCount.subcatchCount.value
        for i in range(self.get_ProjectSize(subcatchment_count)):
            NAME = ctypes.create_string_buffer(46)
            LEN = ctypes.c_int(46)
            ErrNo1 = self._getIDs(
                self.smoapi,
                tka.SMO_elementType.SM_subcatch.value,
                i,
                ctypes.byref(NAME),
                ctypes.byref(LEN),
            )
            if ErrNo1 != 0:
                error_msg = "API ErrNo {0}:{1}".format(
                    ErrNo1.value, tka.DLLErrorKeys[ErrNo1.value]
                )
                raise Exception(error_msg)
            self.SubcatchmentIDs[str(NAME.value)] = i

    def _get_NodeIDs(self):
        """Generates member Element IDs dictionary for Nodes."""
        self.NodeIDs = {}
        node_count = tka.SMO_elementCount.nodeCount.value
        for i in range(self.get_ProjectSize(node_count)):
            NAME = ctypes.create_string_buffer(46)
            LEN = ctypes.c_int(46)
            ErrNo1 = self._getIDs(
                self.smoapi,
                tka.SMO_elementType.SM_node.value,
                i,
                ctypes.byref(NAME),
                ctypes.byref(LEN),
            )
            if ErrNo1 != 0:
                error_msg = "API ErrNo {0}:{1}".format(
                    ErrNo1.value, tka.DLLErrorKeys[ErrNo1.value]
                )
                raise Exception(error_msg)
            self.NodeIDs[str(NAME.value)] = i

    def _get_LinkIDs(self):
        """Generates member Element IDs dictionary for Links."""
        self.LinkIDs = {}
        link_count = tka.SMO_elementCount.linkCount.value
        for i in range(self.get_ProjectSize(link_count)):
            NAME = ctypes.create_string_buffer(46)
            LEN = ctypes.c_int(46)
            ErrNo1 = self._getIDs(
                self.smoapi,
                tka.SMO_elementType.SM_link.value,
                i,
                ctypes.byref(NAME),
                ctypes.byref(LEN),
            )
            if ErrNo1 != 0:
                error_msg = "API ErrNo {0}:{1}".format(
                    ErrNo1.value, tka.DLLErrorKeys[ErrNo1.value]
                )
                raise Exception(error_msg)
            self.LinkIDs[str(NAME.value)] = i

    def _get_PollutantIDs(self):
        """Generates member Element IDs dictionary for Pollutants."""
        self.PollutantIDs = {}
        pollutant_count = tka.SMO_elementCount.pollutantCount.value
        for i in range(self.get_ProjectSize(pollutant_count)):
            NAME = ctypes.create_string_buffer(46)
            LEN = ctypes.c_int(46)
            ErrNo1 = self._getIDs(
                self.smoapi,
                tka.SMO_elementType.SM_sys.value,
                i,
                ctypes.byref(NAME),
                ctypes.byref(LEN),
            )
            if ErrNo1 != 0:
                error_msg = "API ErrNo {0}:{1}".format(
                    ErrNo1.value, tka.DLLErrorKeys[ErrNo1.value]
                )
                raise Exception(error_msg)
            self.PollutantIDs[str(NAME.value)] = i

    def get_IDs(self, SMO_elementIDType):
        """
        Returns List Type of Element IDs.

        :param int SMO_elementCount: element ID type :doc:`/keyrefs`

        :return: list ordered List of IDs
        :rtype: list

        Examples:

        >>> OutputFile = SWMMBinReader()
        >>> OutputFile.OpenBinFile("outputfile.out")
        >>> Test.get_IDs(SM_subcatch)
        >>> ['S3', 'S2', 'S1']
        >>> Test.get_IDs(SM_node)
        >>> ['J4', 'J1', 'J2', 'J3']
        >>> Test.get_IDs(SM_link)
        >>> ['C3', 'C2', 'C1']
        """
        if SMO_elementIDType == tka.SMO_elementType.SM_subcatch.value:
            if not hasattr(self, "SubcatchmentIDs"):
                self._get_SubcatchIDs()
            IDlist = self.SubcatchmentIDs.keys()
        elif SMO_elementIDType == tka.SMO_elementType.SM_node.value:
            if not hasattr(self, "NodeIDs"):
                self._get_NodeIDs()
            IDlist = self.NodeIDs.keys()
        elif SMO_elementIDType == tka.SMO_elementType.SM_link.value:
            if not hasattr(self, "LinkIDs"):
                self._get_LinkIDs()
            IDlist = self.LinkIDs.keys()
        elif SMO_elementIDType == tka.SMO_elementType.SM_sys.value:
            if not hasattr(self, "PollutantIDs"):
                self._get_PollutantIDs()
            IDlist = self.PollutantIDs.keys()
        else:
            error_msg = "SMO_elementType: {} Outside Valid Types".format(
                tka.SMO_elementType
            )
            raise Exception(error_msg)
            return 0

        # Do not sort lists
        return IDlist

    def get_Units(self, unit):
        """
        Returns flow units and Concentration.

        :param int SMO_unit: element ID type :doc:`/keyrefs`

        :return: Unit Type
        :rtype: str

        Examples:

        >>> OutputFile = SWMMBinReader()
        >>> OutputFile.OpenBinFile("outputfile.out")
        >>> OutputFile.get_Units(flow_rate)
        >>> 'CFS'
        """
        FlowUnitsType = [
            "CFS",  # cubic feet per second
            "GPM",  # gallons per minute
            "MGD",  # million gallons per day
            "CMS",  # cubic meters per second
            "LPS",  # liters per second
            "MLD",  # million liters per day
        ]

        ConcUnitsType = [
            "mg",  # Milligrams / L
            "ug",  # Micrograms / L
            "COUNT",  # Counts / L
        ]

        x = ctypes.c_int()
        ErrNo1 = self._getUnits(self.smoapi, unit, ctypes.byref(x))
        if ErrNo1 != 0:
            error_msg = "API ErrNo {0}:{1}".format(ErrNo1, tka.DLLErrorKeys[ErrNo1])
            raise Exception(error_msg)
        if unit == tka.SMO_unit.flow_rate.value:
            return FlowUnitsType[x.value]
        elif unit == tka.SMO_unit.concentration.value:
            return ConcUnitsType[x.value]
        else:
            error_msg = "SMO_unit: {} Outside Valid Types".format(unit)
            raise Exception(error_msg)

    def get_Times(self, SMO_timeElementType):
        """
        Returns report and simulation time related parameters.

        :param int SMO_timeElementType: element ID type :doc:`/keyrefs`

        :return: Report Step (seconds) or Number of Time Steps
        :rtype: int

        Examples:

        >>> OutputFile = SWMMBinReader()
        >>> OutputFile.OpenBinFile("outputfile.out")
        >>> OutputFile.get_Times(reportStep)
        >>> 300
        """
        timeElement = ctypes.c_int()
        ErrNo1 = self._getTimes(
            self.smoapi, SMO_timeElementType, ctypes.byref(timeElement)
        )
        if ErrNo1 != 0:
            error_msg = "API ErrNo {0}:{1}".format(ErrNo1, tka.DLLErrorKeys[ErrNo1])
            raise Exception(error_msg)
        return timeElement.value

    def _get_StartTimeSWMM(self):
        """Returns the simulation start datetime as double."""
        StartTime = ctypes.c_double()
        ErrNo1 = self._getStartTime(self.smoapi, ctypes.byref(StartTime))
        if ErrNo1 != 0:
            error_msg = "API ErrNo {0}:{1}".format(ErrNo1, tka.DLLErrorKeys[ErrNo1])
            raise Exception(error_msg)
        return StartTime.value

    def get_StartTime(self):
        """
        Uses SWMM5 Conversion Functions to Pull DateTime String.

        This converts to Python datetime format.

        :return: Simulation start time.
        :rtype: datetime

        Examples:

        >>> OutputFile = SWMMBinReader()
        >>> OutputFile.OpenBinFile("outputfile.out")
        >>> OutputFile.get_StartTime()
        >>> datetime.datetime(2016,10,4,12,4,0)
        """
        _StartTime = self._get_StartTimeSWMM()
        _date = int(_StartTime)
        _time = _StartTime - _date

        # Pull Date String
        DateStr = ctypes.create_string_buffer(50)
        self.SWMMdateToStr(ctypes.c_double(_date), ctypes.byref(DateStr))
        DATE = DateStr.value

        # Pull Time String
        TimeStr = ctypes.create_string_buffer(50)
        self.SWMMtimeToStr(ctypes.c_double(_time), ctypes.byref(TimeStr))
        TIME = TimeStr.value
        DTime = datetime.strptime(DATE + " " + TIME, "%Y-%b-%d %H:%M:%S")
        return DTime

    def get_TimeSeries(self):
        """
        Gets simulation start time and builds timeseries array based step.

        :return: Simulation time series.
        :rtype: list of datetime

        Examples:

        >>> OutputFile = SWMMBinReader()
        >>> OutputFile.OpenBinFile("outputfile.out")
        >>> OutputFile.get_TimeSeries()
        >>> [datetime.datetime(2015, 11, 29, 14, 0),
             datetime.datetime(2015, 11, 29, 14, 1), ...,
             datetime.datetime(2015, 11, 29, 14, 9)]
        """
        return [
            self.get_StartTime()
            + timedelta(seconds=ind * self.get_Times(tka.SMO_time.reportStep.value))
            for ind in range(self.get_Times(tka.SMO_time.numPeriods.value))
        ]

    def get_ProjectSize(self, SMO_elementCount):
        """
        Returns number of elements of a specific element type.

        :param int SMO_elementCount: element ID type :doc:`/keyrefs`

        :return: Number of Objects
        :rtype: int

        Examples:

        >>> OutputFile = SWMMBinReader()
        >>> OutputFile.OpenBinFile("outputfile.out")
        >>> OutputFile.get_ProjectSize(nodeCount)
        >>> 10
        """
        numel = ctypes.c_int()
        ErrNo1 = self._getProjectSize(
            self.smoapi, SMO_elementCount, ctypes.byref(numel)
        )
        if ErrNo1 != 0:
            error_msg = "API ErrNo {0}:{1}".format(ErrNo1, tka.DLLErrorKeys[ErrNo1])
            raise Exception(error_msg)
        return numel.value

    def get_Series(
        self, element_type, SMO_Attribute, IDName=None, TimeStartInd=0, TimeEndInd=-1
    ):
        """
        Get time series results for particular attribute for an object.

        Specify series start and length using TimeStartInd and TimeEndInd
        respectively.

        :param int SMO_elementType: Element type :doc:`/keyrefs`.
        :param int SMO_Attribute: Attribute Type :doc:`/keyrefs`.
        :param str IDName: Element ID name (Default is None for to reach sys
                           variables) (ID Names are case sensitive).
        :param int TimeStartInd: Starting index for the time series data
                                 period (default is 0).
        :param int TimeEndInd: Array index for the time series data period
                               (defualt is -1 for end).

        :return: data series
        :rtype: list

        Examples:

        >>> OutputFile = SWMMBinReader()
        >>> OutputFile.OpenBinFile("outputfile.out")
        >>> OutputFile.get_Series(SM_subcatch, runoff_rate, 'S3', 0, 50)
        >>> [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, .... 0.0]

        >>> OutputFile.get_Series(SM_node, invert_depth, 'J1', 0, 50)
        >>> [3.908519983291626, 4.6215434074401855, 4.594745635986328,
             4.595311641693115, ...,  4.595311641693115]

        >>> OutputFile.get_Series(SM_link, rainfall_subcatch, 'C2', 0, 50)
        >>> [10.2869873046875, 10.04793643951416, 9.997148513793945,
             10.000744819641113, ..., 10.011372566223145]

        >>> OutputFile.get_Series(SM_sys, rainfall_system, TimeStartInd=0,
                                  TimeEndInd=50)
        >>> [0.017500000074505806, 0.017500000074505806, 0.017500000074505806,
             0.017500000074505806, ..., 0.017500000074505806]
        """
        if TimeEndInd > self.get_Times(tka.SMO_time.numPeriods.value):
            raise Exception("Outside Number of TimeSteps")
        elif TimeEndInd == -1:
            TimeEndInd = 1
            TimeEndInd += self.get_Times(tka.SMO_time.numPeriods.value)
            -TimeEndInd

        sLength = ctypes.c_int()
        ErrNo1 = ctypes.c_int()
        SeriesPtr = self._newOutValueSeries(
            self.smoapi,
            TimeStartInd,
            TimeEndInd,
            ctypes.byref(sLength),
            ctypes.byref(ErrNo1),
        )
        if ErrNo1.value != 0:
            error_msg = "API ErrNo {0}:{1}".format(
                ErrNo1.value, tka.DLLErrorKeys[ErrNo1.value]
            )
            raise Exception(error_msg)

        if element_type == tka.SMO_elementType.SM_subcatch.value:
            if not hasattr(self, "SubcatchmentIDs"):
                self._get_SubcatchIDs()
            ErrNo2 = self._getSubcatchSeries(
                self.smoapi,
                self.SubcatchmentIDs[IDName],
                SMO_Attribute,
                TimeStartInd,
                sLength.value,
                SeriesPtr,
            )
        elif element_type == tka.SMO_elementType.SM_node.value:
            if not hasattr(self, "NodeIDs"):
                self._get_NodeIDs()
            ErrNo2 = self._getNodeSeries(
                self.smoapi,
                self.NodeIDs[IDName],
                SMO_Attribute,
                TimeStartInd,
                sLength.value,
                SeriesPtr,
            )
        elif element_type == tka.SMO_elementType.SM_link.value:
            if not hasattr(self, "LinkIDs"):
                self._get_LinkIDs()
            ErrNo2 = self._getLinkSeries(
                self.smoapi,
                self.LinkIDs[IDName],
                SMO_Attribute,
                TimeStartInd,
                sLength.value,
                SeriesPtr,
            )
        # Add Pollutants Later
        elif element_type == tka.SMO_elementType.SM_sys.value:
            ErrNo2 = self._getSystemSeries(
                self.smoapi, SMO_Attribute, TimeStartInd, sLength.value, SeriesPtr
            )
        else:
            error_msg = "SMO_elementType: {} Outside Valid Types".format(element_type)
            raise Exception(error_msg)

        if ErrNo2 != 0:
            error_msg = "API ErrNo {0}:{1}".format(ErrNo2, tka.DLLErrorKeys[ErrNo2])
            raise Exception(error_msg)

        BldArray = [SeriesPtr[i] for i in range(sLength.value)]
        self._free(SeriesPtr)
        return BldArray

    def get_Attribute(self, element_type, SMO_Attribute, TimeInd):
        """
        Get results for attribute for elements at a specific time.

        :param int SMO_elementType: Element type :doc:`/keyrefs`.
        :param int SMO_Attribute: Attribute Type :doc:`/keyrefs`.
        :param int TimeInd: TimeInd

        :return: data list in order of the IDs of the SMO_elementType
        :rtype: list

        Examples:

        >>> OutputFile = SWMMBinReader()
        >>> OutputFile.OpenBinFile("outputfile.out")
        >>> OutputFile.get_Attribute(SM_subcatch, rainfall_subcatch, 0)
        >>> [0.017500000074505806, 0.017500000074505806, 0.017500000074505806]

        >>> OutputFile.get_Attribute(SM_node, invert_depth, 10)
        >>> [4.596884250640869, 0.720202624797821, 0.6315776705741882,
             0.6312257051467896]

        >>> OutputFile.get_Attribute(SM_link, flow_rate_link, 50)
        >>> [9.00419807434082, 10.011459350585938, 11.020767211914062]
        """
        if TimeInd > self.get_Times(tka.SMO_time.numPeriods.value) - 1:
            raise Exception("Outside Number of TimeSteps")

        aLength = ctypes.c_int()
        ErrNo1 = ctypes.c_int()
        ValArrayPtr = self._newOutValueArray(
            self.smoapi,
            tka.SMO_apiFunction.getAttribute.value,
            element_type,
            ctypes.byref(aLength),
            ctypes.byref(ErrNo1),
        )
        if ErrNo1.value != 0:
            error_msg = "API ErrNo {0}:{1}".format(
                ErrNo1.value, tka.DLLErrorKeys[ErrNo1.value]
            )
            raise Exception(error_msg)

        if element_type == tka.SMO_elementType.SM_subcatch.value:
            ErrNo2 = self._getSubcatchAttribute(
                self.smoapi, TimeInd, SMO_Attribute, ValArrayPtr
            )
        elif element_type == tka.SMO_elementType.SM_link.value:
            ErrNo2 = self._getLinkAttribute(
                self.smoapi, TimeInd, SMO_Attribute, ValArrayPtr
            )
        elif element_type == tka.SMO_elementType.SM_node.value:
            ErrNo2 = self._getNodeAttribute(
                self.smoapi, TimeInd, SMO_Attribute, ValArrayPtr
            )
        # Add Pollutants Later
        else:
            error_msg = "SMO_elementType: {} Outside Valid Types".format(element_type)
            raise Exception(error_msg)

        if ErrNo2 != 0:
            error_msg = "API ErrNo {0}:{1}".format(ErrNo2, tka.DLLErrorKeys[ErrNo2])
            raise Exception(error_msg)

        BldArray = [ValArrayPtr[i] for i in range(aLength.value)]
        self._free(ValArrayPtr)
        return BldArray

    def get_Result(self, element_type, TimeInd, IDName=None):
        """
        For a element ID at given time, get all attributes.

        :param int SMO_elementType: Element type :doc:`/keyrefs`.
        :param int TimeInd: Time Index
        :param int IDName: IDName (default None for System Variables)

        Examples:

        >>> OutputFile = SWMMBinReader()
        >>> OutputFile.OpenBinFile("outputfile.out")
        >>> OutputFile.get_Result(SM_subcatch,3000,'S3')
        >>> [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0]

        >>> OutputFile.get_Result(SM_node,3000,'J1')
        >>> [4.594789505004883, 25.322790145874023, 0.0, 9.000000953674316,
             9.000000953674316, 0.0]

        >>> OutputFile.get_Result(SM_link,9000,'C3')
        >>> [11.0, 0.6312892436981201, 12.93112564086914,
             185.72474670410156, 0.270773708820343]

        >>> OutputFile.get_Result(SM_sys,3000,'S3')
        >>> [70.0, 0.0, 0.0, 0.0, 0.0, 8.0, 0.0, 0.0, 3.0, 11.0, 0.0,
             11.000021934509277, 532.2583618164062, 0.0, 0.0]
        """
        if TimeInd > self.get_Times(tka.SMO_time.numPeriods.value) - 1:
            raise Exception("Outside Number of TimeSteps")

        alength = ctypes.c_int()
        ErrNo1 = ctypes.c_int()
        ValArrayPtr = self._newOutValueArray(
            self.smoapi,
            tka.SMO_apiFunction.getResult,
            element_type,
            ctypes.byref(alength),
            ctypes.byref(ErrNo1),
        )

        if element_type == tka.SMO_elementType.SM_subcatch.value:
            if not hasattr(self, "SubcatchmentIDs"):
                self._get_SubcatchIDs()
            ErrNo2 = self._getSubcatchResult(
                self.smoapi, TimeInd, self.SubcatchmentIDs[IDName], ValArrayPtr
            )
        elif element_type == tka.SMO_elementType.SM_node.value:
            if not hasattr(self, "NodeIDs"):
                self._get_NodeIDs()
            ErrNo2 = self._getNodeResult(
                self.smoapi, TimeInd, self.NodeIDs[IDName], ValArrayPtr
            )
        elif element_type == tka.SMO_elementType.SM_link.value:
            if not hasattr(self, "LinkIDs"):
                self._get_LinkIDs()
            ErrNo2 = self._getLinkResult(
                self.smoapi, TimeInd, self.LinkIDs[IDName], ValArrayPtr
            )
        # Add Pollutants Later
        elif element_type == tka.SMO_elementType.SM_sys.value:
            ErrNo2 = self._getSystemResult(self.smoapi, TimeInd, ValArrayPtr)
        else:
            error_msg = "SMO_elementType: {} Outside Valid Types".format(element_type)
            raise Exception(error_msg)

        ErrNo2  # FIXME: is this not used?
        BldArray = [ValArrayPtr[i] for i in range(alength.value)]
        self._free(ValArrayPtr)
        return BldArray


if __name__ in "__main__":
    # Run Tests

    # Open
    path = r"C:\PROJECTCODE\pyswmm\pyswmm\tests\data\model_node_inflows.out"
    Test = SWMMBinReader()
    Test.OpenBinFile(path)

    # Get IDs
    print("\nProject Element ID Info")
    print(Test.get_IDs(tka.SMO_elementType.SM_subcatch.value))
    print(Test.get_IDs(tka.SMO_elementType.SM_node.value))
    print(Test.get_IDs(tka.SMO_elementType.SM_link.value))

    print("\nGet Units")
    print("flow_rate: {}".format(Test.get_Units(tka.SMO_unit.flow_rate.value)))

    # Get Project Size
    print("\nProject Size Info")
    print(
        "Subcatchments: {}".format(
            Test.get_ProjectSize(tka.SMO_elementCount.subcatchCount.value)
        )
    )
    print(
        "Nodes: {}".format(Test.get_ProjectSize(tka.SMO_elementCount.nodeCount.value))
    )
    print(
        "Links: {}".format(Test.get_ProjectSize(tka.SMO_elementCount.linkCount.value))
    )
    print(
        "Pollutants: {}".format(
            Test.get_ProjectSize(tka.SMO_elementCount.pollutantCount.value)
        )
    )

    # Project Time Steps
    print("\nProject Time Info")
    print("Report Step: {}".format(Test.get_Times(tka.SMO_time.reportStep.value)))
    print("Periods: {}".format(Test.get_Times(tka.SMO_time.numPeriods.value)))

    # Get Time Series
    print("\nGet Time Series")
    TimeSeries = Test.get_TimeSeries()
    print(TimeSeries[:10])

    # Get Series
    print("\nSeries Tests")
    SubcSeries = Test.get_Series(
        tka.SMO_elementType.SM_subcatch.value,
        tka.SMO_systemAttribute.runoff_rate,
        "S3",
        0,
        50,
    )
    print(SubcSeries)
    NodeSeries = Test.get_Series(
        tka.SMO_elementType.SM_node.value,
        tka.SMO_systemAttribute.invert_depth,
        "J1",
        0,
        50,
    )
    print(NodeSeries)
    LinkSeries = Test.get_Series(
        tka.SMO_elementType.SM_link.value,
        tka.SMO_systemAttribute.rainfall_subcatch,
        "C2",
        0,
        50,
    )
    print(LinkSeries)
    SystSeries = Test.get_Series(
        tka.SMO_elementType.SM_sys.value,
        tka.SMO_systemAttribute.rainfall_system.value,
        TimeStartInd=0,
        TimeEndInd=50,
    )
    print(SystSeries)

    # Get Attributes
    print("\nAttributes Tests")
    # Check Values.. Might be issue here
    SubcAttributes = Test.get_Attribute(
        tka.SMO_elementType.SM_subcatch.value,
        tka.SMO_systemAttribute.rainfall_subcatch.value,
        0,
    )
    print(SubcAttributes)
    NodeAttributes = Test.get_Attribute(
        tka.SMO_elementType.SM_node.value,
        tka.SMO_systemAttribute.invert_depth.value,
        10,
    )
    print(NodeAttributes)
    LinkAttributes = Test.get_Attribute(
        tka.SMO_elementType.SM_link.value,
        tka.SMO_systemAttribute.flow_rate_link.value,
        50,
    )
    print(LinkAttributes)

    # Get Results
    print("\nResult Tests")
    SubcResults = Test.get_Result(tka.SMO_elementType.SM_subcatch.value, 3000, "S3")
    print(SubcResults)
    NodeResults = Test.get_Result(tka.SMO_elementType.SM_node.value, 3000, "J1")
    print(NodeResults)
    LinkResults = Test.get_Result(tka.SMO_elementType.SM_link.value, 9000, "C3")
    print(LinkResults)
    SystResults = Test.get_Result(tka.SMO_elementType.SM_sys.value, 3000, "S3")
    print(SystResults)

    # Close Output File
    Test.CloseBinFile()
    help(SWMMBinReader)
